# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import time
import math
import random
import numpy
import dismod_at
import at_cascade
"""
{xrst_begin csv.simulate}
{xrst_spell
  cv
  meas
  pini
  sincidence
  std
}

Simulate A Cascade Data Set
###########################

Prototype
*********
{xrst_literal
   # BEGIN_SIMULATE
   # END_SIMULATE
}

sim_dir
*******
This string is the directory name where the csv files
are located.

Example
*******
:ref:`csv.simulate_xam-name`


Input Files
***********

option_sim.csv
==============
This csv file has two columns,
one called ``name`` and the other called ``value``.
The rows of this table are documented below by the name column.
If an option name does not appear, the corresponding
default value is used for the option.
The final value for each of the options is reported in the file
:ref:`csv.simulate@Output Files@option_sim_out.csv` .
Because each option has a default value,
new option are added in such a way that
previous option_sim.csv files are still valid.

absolute_covariates
-------------------
This is a space separated list of the names of the absolute covariates.
The reference value for an absolute covariate is always zero.
(The reference value for a relative covariate is its average for the
location that is being fit.)
The default value for *absolute_covariates* is the empty string; i.e.,
there are no absolute covariates.

absolute_tolerance
------------------
This float is the absolute error tolerance for the integrator.
It determines the accuracy of
:ref:`csv.simulate@Output Files@data_sim.csv@meas_mean` for
:ref:`integrand <csv.simulate@Input Files@simulate.csv@integrand_name>`
that require the ODE; e.g.,
prevalence requires the ODE and Sincidence does not.
The default value for this option is 1e-5.

float_precision
---------------
This integer is the number of decimal digits of precision to
include for float values in the output csv files.
The default value for this option is 4.

integrand_step_size
-------------------
This float is the step size in age and time used to approximate
integrand averages from age_lower to age_upper
and time_lower to time_upper (in data_sim.csv).
It must be greater than zero.
The default value for this option is 5.0.

random_depend_sex
-----------------
If :ref:`csv.simulate@Input Files@option_sim.csv@new_random_effects` is
false, this option is not used.
Otherwise if this boolean is true, the random effects depend on sex.
if it is false, for each *node_name* and *rate*, the random effect for
``female`` and ``male`` will be equal; see
:ref:`csv.simulate@random_effect.csv` .
The default value for this option is false.

new_random_effects
------------------
If this boolean is true, a new set of random effects is generated and
:ref:`csv.simulate@random_effect.csv` is an output file.
Otherwise random_effect.csv is an input file.
The default value for this boolean is true.

random_seed
-----------
This integer is used to seed the random number generator.
The default value for this option is
{xrst_code py}
   random_seed = int( time.time() )
{xrst_code}

std_random_effects_rate
-----------------------
If :ref:`csv.simulate@Input Files@option_sim.csv@new_random_effects` is
false, this option is not used.
Otherwise, this float is the standard deviation of the random effects
for the corresponding *rate* where *rate* is pini, iota, rho, or chi.
The effects are in log of rate space, so this standard deviation
is also in log of rate space.
Hence only the rates that appear in
:ref:`csv.simulate@Input Files@no_effect_rate.csv`
have an effect (the other random effects multiply zero).
The default value for this option is 0.0; i.e.,
there are no random effects for the corresponding rate.

trace
-----
If this boolean is true, a trace will be printed during the simulation.
This will show that the simulation is making progress and is useful
for cases where there is a lot of data to simulate.
The default value for this boolean is true.

-----------------------------------------------------------------------------

node.csv
========
This csv file defines the node tree.
It has the columns documented below.

node_name
---------
This string is a name describing the node in a way that is easy for a human to
remember. It be unique for each row.

parent_name
-----------
This string is the node name corresponding to the parent of this node.
The root node of the tree has an empty entry for this column.
If a node is a parent, it must have at least two children.
This avoids fitting the same location twice as one goes from parent
to child nodes.

-----------------------------------------------------------------------------

covariate.csv
=============
This csv file specifies the value of omega and the covariates.
For each node_name it has a rectangular grid in age and time.
In addition, the rectangular grid is the same for nodes.

node_name
---------
This string identifies the node, in node.csv, corresponding to this row.

sex
---
This identifies which sex this row corresponds to.
The sex values ``female`` and ``male`` must appear in this table.
The sex value ``both`` does not appear.

age
---
This float is the age, in years,  corresponding to this row.

time
----
This float is the time, in years, corresponding to this row.

omega
-----
This float is the value of omega (other cause mortality) for this row.
Often other cause mortality is approximated by all cause mortality.
Omega is a rate that is assumed to be know ahead of time
and hence it is specified together with the covariates.

covariate_name
--------------
Except for node_name, sex, age. time, and omega,
the columns of this file are covariates.
The header row specifies the *covariate_name* for a column
and the other rows are floats containing the corresponding covariate value.
The option_sim.csv
:ref:`csv.simulate@Input Files@option_sim.csv@absolute_covariates`
specifies which covariates are absolute.
All the others are :ref:`relative covariates<glossary@Relative Covariate>`.
Note that omega and sex are not referred to as covariates for this simulation.

-----------------------------------------------------------------------------

no_effect_rate.csv
==================
This csv file specifies the grid points at which each rate is modeled
during a simulation. For each rate_name it has a
:ref:`csv.module@Notation@Rectangular Grid` in age and time.
These are no-effect rates; i.e., the rates without
the random and covariate effects.
Covariate multipliers that are constrained to zero during the fitting
can be used to get variation between nodes in the
no-effect rates corresponding to the fit.

rate_name
---------
This string is ``iota``, ``rho``, ``chi``, or ``pini`` and specifies the rate.
If one of these rates does not appear, it is modeled as always zero.
Other cause mortality ``omega`` is specified in
:ref:`csv.simulate@Input Files@covariate.csv` .

age
---
This float is the age, in years,  corresponding to this row.

time
----
This float is the time, in years, corresponding to this row.

rate_truth
----------
This float is the no-effect rate value for all the nodes.
It is used to simulate the data.
As mentioned, above knocking out covariate multipliers can be
used to get variation in the no-effect rates that correspond to the fit.
If *rate_name* is ``pini``, *rate_truth*  should be constant w.r.t *age*
(because it is prevalence at age zero).

-----------------------------------------------------------------------------

multiplier_sim.csv
==================
This csv file provides information about the covariate multipliers.
Each row of this file, except the header row, corresponds to a
different multiplier. The multipliers are constant in age and time.

multiplier_id
-------------
is an :ref:`csv.module@Notation@Index Column` for multiplier_sim.csv.

rate_name
---------
This string is ``iota``, ``rho``, ``chi``, or ``pini`` and specifies
which rate this covariate multiplier is affecting.

covariate_or_sex
----------------
If this is ``sex`` it specifies that this multiplier multiples
the sex values where
{xrst_code py}"""
sex_covariate_value = { 'female' : -0.5,  'both' : 0.0, 'male' : +0.5 }
"""{xrst_code}
female = -0.5, male = +0.5, and both = 0.0.
Otherwise this is one of the covariate names in the covariate.csv file
and specifies which covariate value is being multiplied.

multiplier_truth
----------------
This is the value of the covariate multiplier used to simulate the data.


-----------------------------------------------------------------------------

simulate.csv
============
This csv file specifies the simulated data set
with each row corresponding to one data point.

simulate_id
-----------
is an :ref:`csv.module@Notation@Index Column` for simulate.csv.

integrand_name
--------------
This string is a dismod_at integrand; e.g. ``Sincidence``.

node_name
---------
This string identifies the node corresponding to this data point.

sex
---
This string is the sex for this data point.

age_lower
---------
This float is the lower age limit for this data row.

age_upper
---------
This float is the upper age limit for this data row.

time_lower
----------
This float is the lower time limit for this data row.

time_upper
----------
This float is the upper time limit for this data row.

meas_std_cv
-----------
This float is the coefficient of variation for the measurement noise
for this data row;
see :ref:`csv.simulate@Output Files@data_sim.csv@meas_std` .

meas_std_min
------------
This float is the minimum value for the standard deviation
of the measurement noise for this data row;
see :ref:`csv.simulate@Output Files@data_sim.csv@meas_std` .

------------------------------------------------------------------------------

random_effect.csv
*****************
This file reports the random effect for each node, rate and sex.
If :ref:`csv.simulate@Input Files@option_sim.csv@new_random_effects`
is true (false) , this an input (output) file.
Only the rate names that appear in
:ref:`csv.simulate@Input Files@no_effect_rate.csv@rate_name`
are included in random_effect.csv .
(Random effect for rates not in no_effect_rate.csv have no effect.)

node_name
=========
This string identifies the row in :ref:`csv.simulate@Input Files@node.csv`
that this row corresponds to.
All of the nodes in the node table are present in this file.

rate_name
=========
This is a string and is one of the
For each :ref:`csv.simulate@Input Files@no_effect_rate.csv@rate_name`
in the no_effect rate table,
All of the rates in the no_effect rate table are present in this file.

sex
===
This identifies which sex the random effect corresponds to.
The sex values ``female`` and ``male`` will appear and
``both`` will not appear.

random_effect
=============
This float value is the random effect for the specified node, rate, and sex.
If new_random_effects is true and
:ref:`csv.simulate@Input Files@option_sim.csv@random_depend_sex`  is false,
the value in this column will not depend on the value in the sex column.

Discussion
==========
1. For a given parent node, rate, and sex,
   the sum of the random effects with respect to the child nodes is zero.
2. All the random effects for the root node are set to zero
   (the root node does not have a parent node).

-----------------------------------------------------------------------------

Output Files
************

option_sim_out.csv
==================
This is a copy of :ref:`csv.simulate@Input Files@option_sim.csv`
with the default filled in for missing values.

data_sim.csv
============
This contains the simulated data.
It is created during a simulate command
and has the following columns:

simulate_id
-----------
This integer identifies the row in the simulate.csv
corresponding to this row in data_sim.csv.
This is an :ref:`csv.module@Notation@Index Column`
for simulate.csv and data_sim.csv.

meas_mean
---------
This float is the mean value for the measurement.
This is the model value without any measurement noise.
It corresponds to the simulation value for all
the model variables and covariates.
We refer to this as the *true value for the average integrand*
even when we have model miss-specification;
i.e., when the set of model variables or covariates in
:ref:`csv.simulate-name` is different from the set in :ref:`csv.fit-name` .

meas_std
--------
This float is the measurement standard deviation for the simulated
data point. This standard deviation is before censoring and given by

|  *meas_std* = max ( *meas_std_min* , *meas_std_cv* * *meas_mean* )

where
:ref:`csv.simulate@Input Files@simulate.csv@meas_std_min`
is the minimum measure standard deviation,
and :ref:`csv.simulate@Input Files@simulate.csv@meas_std_cv`
is the coefficient of variation for the measurement noise.

meas_value
----------
This float is the simulated measured value.
The data will be generated with a normal distribution
that has mean *meas_mean* and standard deviation *meas_std* .
If the resulting measurement value would be less than zero,
the value zero is used; i.e.,
a censored normal is used to simulate the data.


covariate_name
--------------
For each :ref:`csv.simulate@Input Files@covariate.csv@covariate_name`
there is a column with this name in simulate.csv.
The values in these columns are floats
corresponding to the covariate value at the mid point of the ages and time
intervals for this data point. This value is obtained using
bilinear interpolation of the covariate values in covariate.csv.
The interpolate is extended as constant in age (time) for points
outside the age rage (time range) in the covariate.csv file.

{xrst_end csv.simulate}
"""
# ----------------------------------------------------------------------------
# set_global_option_value
#
# Sets global global_option_value to dict representation of option_sim.csv
#
# sim_dir
# is the directory where the input csv files are located.
#
# option_table :
# is the list of dict corresponding to option_sim.csv
#
# option_sim_out.csv
# As a side effect, this routine write a copy of the option table
# with the default values filled in.
#
# global_option_value[name] :
# is the option value corresponding the specified option name.
# Here name is a string and value
# has been converted to its corresponding type.
#
global_option_value = None
def set_global_option_value(sim_dir, option_table) :
   global global_option_value
   assert type(global_option_value) == dict or global_option_value == None
   assert type(option_table) == list
   if len(option_table) > 0 :
      assert type( option_table[0] ) == dict
   #
   # option_default
   random_seed = int( time.time() )
   # BEGIN_SORT_THIS_LINE_PLUS_2
   option_default  = {
      'absolute_covariates'              : (str, None)         ,
      'absolute_tolerance'               : (float, 1e-5)       ,
      'float_precision'                  : (int,   4)          ,
      'integrand_step_size'              : (float, 5.0)        ,
      'new_random_effects'               : (bool,  True)       ,
      'random_depend_sex'                : (bool,  False)      ,
      'random_seed'                      : (int, random_seed)  ,
      'std_random_effects_chi'           : (float, 0.0)        ,
      'std_random_effects_iota'          : (float, 0.0)        ,
      'std_random_effects_pini'          : (float, 0.0)        ,
      'std_random_effects_rho'           : (float, 0.0)        ,
      'trace'                            : (bool,  True)       ,
   }
   # END_SORT_THIS_LINE_MINUS_2
   #
   # global_option_value
   line_number      = 0
   global_option_value = dict()
   for row in option_table :
      line_number += 1
      name         = row['name']
      if name in global_option_value :
         msg  = f'csv.simulate: Error: line {line_number} in option_sim.csv\n'
         msg += f'the name {name} appears twice in this table'
         assert False, msg
      if not name in option_default :
         msg  = f'csv.simulate: Error: line {line_number} in option_sim.csv\n'
         msg += f'{name} is not a valid option name'
         assert False, msg
      (option_type, default) = option_default[name]
      value                  = row['value']
      if value == '' :
         option_value[name] = None
      elif option_type == bool :
         if value not in [ 'true', 'false' ] :
            msg  = f'csv.simulate: Error: line {line_number} in '
            msg += 'option_sim.csv\n'
            msg += f'The value for {name} is not true or false'
            assert False, msg
         global_option_value[name] = value == 'true'
      else :
         global_option_value[name] = option_type( value )
   #
   # global_option_value
   for name in option_default :
      if name not in global_option_value :
         (option_type, default) = option_default[name]
         global_option_value[name] = default
   #
   # option_sim_out.csv
   table = list()
   for name in global_option_value :
      value = global_option_value[name]
      if type(value) == bool :
         if value :
            value = 'true'
         else :
            value = 'false'
      row = { 'name' : name , 'value' : value }
      table.append(row)
   file_name = f'{sim_dir}/option_sim_out.csv'
   at_cascade.csv.write_table(file_name, table)
   #
   assert type(global_option_value) == dict
# ----------------------------------------------------------------------------
# get_parent_node_dict
#
# parent_node_dict[node_name]:
# The is the name of the node that is the parent of node_node.
# The keys and values in this dictionary are strings.
#
# child_list_node[node_name]:
# A ``list`` of nodes that has all the children of node_name
# where node_name is an ``str`` .
#
# parent_node_dict, child_list_node =
def get_parent_node_dict( node_table ) :
   #
   # parent_node_dict, count_children, root_node_name
   line_number = 0
   parent_node_dict  = dict()
   child_list_node   = dict()
   root_node_name    = None
   for row in node_table :
      line_number += 1
      node_name                  = row['node_name']
      parent_name                = row['parent_name']
      child_list_node[node_name] = list()
      if node_name in parent_node_dict :
         msg  = f'csv.simulate: Error: line {line_number} in node.csv\n'
         msg += f'node_name {node_name} appears twice'
         assert False, msg
      parent_node_dict[node_name] = parent_name
      #
      if parent_name == '' :
         if root_node_name != None :
            msg  = f'csv.simulate: Error: line {line_number} in node.csv\n'
            msg += 'one and only one node should have no parent\n'
            msg += 'node {node_name} and {root_node_name} have no parent'
            assert False, msg
         root_node_name = node_name
   if root_node_name == None :
      msg  = f'csv.simulate: Error in node.csv\n'
      msg += 'there is no root node; i.e.,  node with no parent node'
      assert False, msg
   #
   # child_list_node
   line_number    = 0
   for row in node_table :
      line_number += 1
      node_name    = row['node_name']
      parent_name  = row['parent_name']
      if parent_name != '' :
         if parent_name not in child_list_node :
            msg  = f'csv.simulate: Error: line {line_number} in node.csv\n'
            msg += f'parent_name {parent_name} is not a valid node_name'
            assert False, msg
         else :
            child_list_node[parent_name].append(node_name)
   #
   # check number of children
   for parent_name in child_list_node :
      if len( child_list_node[parent_name] ) == 1 :
         msg  = 'csv.simulate: Error in node.csv\n'
         msg += f'the parent_name {parent_name} appears once and only once'
         assert False, msg
   #
   # check that no node is and ancestor of itself
   for node_name in parent_node_dict :
      ancestor_set = { node_name }
      parent_name  = parent_node_dict[node_name]
      while parent_name != '' :
         if parent_name in ancestor_set :
            msg  = 'csv.simulate: Error in node_table.csv\n'
            msg += f'node {parent_name} is an ancestor of itself'
            assert False, msg
         ancestor_set.add(parent_name)
         parent_name = parent_node_dict[parent_name]
   #
   return parent_node_dict, child_list_node
# ----------------------------------------------------------------------------
# get_spline_no_effect_rate
#
# spline = spline_no_effect_rate[rate_name] :
# 1. rate_name is any of the rate names in the no_effect_rate table.
# 2. value = spline(age, time) evaluates the no_effect spline
#    for rate_name at the specified age and time where age, time, value
#    are all floats.
#
# no_effect_rate_table:
# is the table corresponding to no_effect_rate.csv
#
# spline_no_effect_rate =
def get_spline_no_effect_rate(no_effect_rate_table) :
   #
   # age_set, time_set
   rate_row_list   = dict()
   for row in no_effect_rate_table :
      rate_name    = row['rate_name']
      age          = float( row['age'] )
      time         = float( row['time'] )
      rate_truth   = float( row['rate_truth'] )
      #
      if rate_name not in rate_row_list :
         rate_row_list[rate_name] = list()
      rate_row_list[rate_name].append( row )
   #
   # spline_no_effect_rate
   spline_no_effect_rate = dict()
   #
   # rate_name
   for rate_name in rate_row_list :
      #
      # age_grid, time_grid, spline_dict
      age_grid, time_grid, spline_dict = at_cascade.bilinear(
         table  = rate_row_list[rate_name],
         x_name = 'age',
         y_name = 'time',
         z_list = [ 'rate_truth' ]
      )
      #
      if spline_dict == None :
         msg  = 'csv.simulate: Error in no_effect_rate.csv\n'
         msg += 'rate_name = {rate_name}\n'
         msg += 'Expected following rectangular grid:\n'
         msg += f'age_grid  = {age_grid}\n'
         msg += f'time_grid = {time_grid}'
         assert False, msg
      #
      # spline_no_effect_rate
      spline_no_effect_rate[rate_name]= spline_dict['rate_truth']
   #
   return spline_no_effect_rate
# ----------------------------------------------------------------------------
# multiplier_list_rate[rate_name]:
# list of rows in multiplier_sim_table that have the specified rate name
# were rate_name is a string.
#
# multiplier_sim_table:
# is the table corresponding to multiplier_sim.csv.
#
# multiplier_list_rate =
def get_multiplier_list_rate(multiplier_sim_table) :
   #
   # multiplier_list_rate
   multiplier_list_rate = dict()
   for rate_name in [ 'pini', 'iota', 'rho', 'chi' ] :
      multiplier_list_rate[rate_name] = list()
   for row in multiplier_sim_table :
      rate_name = row['rate_name']
      multiplier_list_rate[rate_name].append(row)
   return multiplier_list_rate
# ----------------------------------------------------------------------------
# eval_spline
#
# spline = spline_node_sex_cov[node_name][sex][cov_name] :
# is a function that evaluates value = spline(age, time) where
# node_name is a node name, sex is 'female' or 'male' (not 'both'),
# cov_name is a covariate name or 'sex',
# age, time, and value are floats
# This is used to compute  covariate values.
#
# node_name
# is the node for which we are evalauting the spline.
#
# sex
# is the node for which we are evalauting the spline.
# This maybe 'female', 'male', or 'both'.
#
# cov_name
# is the covariiate for which we are evaluating the spline.
#
# age
# is the age at which we are evaluating the spline.
#
# time
# is the time at which we are evaluating the spline.
#
def eval_spline(spline_node_sex_cov, node_name, sex, cov_name, age, time) :
   assert type(node_name) == str
   assert type(sex) == str
   assert type(cov_name) == str
   assert type(age) == float
   assert type(time) == float
   assert sex in { 'female', 'male', 'both' }
   if sex == 'both' :
      spline = spline_node_sex_cov[node_name]['female'][cov_name]
      female = spline(age, time)
      spline = spline_node_sex_cov[node_name]['male'][cov_name]
      male   = spline(age, time)
      result = (female + male) / 2.0
   else :
      spline = spline_node_sex_cov[node_name][sex][cov_name]
      result = spline(age, time)
   return result
# ----------------------------------------------------------------------------
# get_rate_fun_dict
#
# rate_fun = rate_fun_dict[rate_name] :
# For each of rate_name in spline_no_effect_rate, value = rate_fun(age, time)
# is the value of the corresponding rate without any of the effects
# where age, time, and value are floats.
#
# parent_node_dict:
# is mapping from each node name to the name of its parent node.
#
# spline_no_effect_rate[rate_name] :
# is a spline that evaluates the specified no_effect rate.
#
# random_effect_node_rate_sex[node_name][rate_name][sex] :
# is the random effect for the specified node, rate, and sex.
# Note that to get the difference from the root node, one has to sum
# the random effect for this node and its ancestors not including the
# root node in the sum.
#
# root_covariate_ref[covariate_name] :
# is the reference for the specified covariate when fitting the root node.
#
# multiplier_list_rate[rate_name] :
# is the list of rows, in the multiplier_sim table, that have rate_name
# equal to the specified rate_name.
#
# node_name :
# The node that rate_fun will compute rates for.
#
# sex :
# The sex that rate_fun will compute rates for.
# This can be female, male or both.
#
# spline = spline_node_sex_cov[node_name][sex][cov_name] :
# is a function that evaluates value = spline(age, time) where
# node_name is a node name, sex is 'female' or 'male' (not 'both'),
# cov_name is a covariate name or 'sex',
# age, time, and value are floats
# This is used to compute  covariate values.
#
# rate_fun_dict =
def get_rate_fun_dict(
   parent_node_dict            ,
   spline_no_effect_rate       ,
   random_effect_node_rate_sex ,
   root_covariate_ref          ,
   multiplier_list_rate        ,
   node_name                   ,
   sex                         ,
   spline_node_sex_cov         ,
) :
   assert sex in { 'female', 'male', 'both' }
   # -----------------------------------------------------------------------
   # rate_fun
   def rate_fun(age, time, rate_name) :
      #
      # no_effect_rate
      spline         = spline_no_effect_rate[rate_name]
      no_effect_rate = spline(age, time)
      #
      # effect
      # random effects
      effect         = 0.0
      ancestor_node  = node_name
      while ancestor_node != '' :
         if sex == 'both' :
            temp = random_effect_node_rate_sex[ancestor_node][rate_name]
            term = ( temp['female'] + temp['male'] ) / 2.0
         else :
            term = random_effect_node_rate_sex[ancestor_node][rate_name][sex]
         effect += term
         ancestor_node = parent_node_dict[ancestor_node]
         if ancestor_node == '' :
            assert term == 0.0
      #
      # effect
      # covariate and sex effects
      for row in multiplier_list_rate[rate_name] :
         assert row['rate_name'] == rate_name
         #
         covariate_or_sex = row['covariate_or_sex']
         if covariate_or_sex == 'sex' :
            difference = sex_covariate_value[sex]
         else :
            covariate_name = covariate_or_sex
            covariate      = eval_spline(
               spline_node_sex_cov, node_name, sex, covariate_name, age, time
            )
            reference      = root_covariate_ref[covariate_name]
            difference     = covariate - reference
         effect    += float( row['multiplier_truth'] ) * difference
      #
      # rate
      rate = math.exp(effect) * no_effect_rate
      return rate
   # -----------------------------------------------------------------------
   # omega_fun
   def omega_fun(age, time) :
      omega = eval_spline(
         spline_node_sex_cov, node_name, sex, 'omega', age, time
      )
      return omega
   #
   rate_fun_dict = dict()
   rate_fun_dict['omega'] = lambda age, time : omega_fun(age, time)
   #
   # cannot loop over rate name because it does not bind on assignment.
   if 'pini' in  spline_no_effect_rate :
      rate_fun_dict['pini'] = \
         lambda age, time : rate_fun(age, time, 'pini')
   if 'iota' in  spline_no_effect_rate :
      rate_fun_dict['iota'] = \
         lambda age, time : rate_fun(age, time, 'iota')
   if 'rho' in  spline_no_effect_rate :
      rate_fun_dict['rho'] = \
         lambda age, time : rate_fun(age, time, 'rho')
   if 'chi' in  spline_no_effect_rate :
      rate_fun_dict['chi'] = \
         lambda age, time : rate_fun(age, time, 'chi')
   #
   #
   return rate_fun_dict
# ----------------------------------------------------------------------------
# average_interand_grid
#
# grid['age'] :
# is a list of age point starting at age_lower, ending at age_upper
# and such that the difference between points is less than or equal
# step_size.
#
# grid['time'] :
# is a list of time point starting at time_lower, ending at time_upper
# and such that the difference between points is less than or equal
# step_size.
#
# step_size :
# The maximum allowable distance between grid point.
# This must be grater than zero
#
# grid =
def average_integrand_grid(
   step_size, age_lower, age_upper, time_lower, time_upper
) :
   #
   # age_grid
   if age_lower == age_upper :
      age_grid = [ age_lower ]
   else :
      n_age    = int( (age_upper - age_lower) / step_size) + 1
      dage     = (age_upper - age_lower) / n_age
      age_grid = [ age_lower + i * dage for i in range(n_age+1) ]
   #
   # time_grid
   if time_lower == time_upper :
      time_grid = [ time_lower ]
   else :
      n_time    = int( (time_upper - time_lower) / step_size) + 1
      dtime     = (time_upper - time_lower) / n_time
      time_grid = [ time_lower + i * dtime for i in range(n_time+1) ]
   # grid
   grid = { 'age' : age_grid , 'time' : time_grid }
   #
   return grid
# ----------------------------------------------------------------------------
# read_random_effect_node_rate_sex
#
# random_effect_node_rate_sex[node_name][rate_name][sex] :
# is the simulated random effect for the corresponding node, sex and rate.
# For a given node_name, sex and rate_name, the sum of
#  random_effect_node_rate_sex[child_node][rate_name][sex]
# is zero where child_node ranges over values in child_list_node[node_name]
# All the random effects for the root node are zero
# (prent_node_dict[root_node_name] == '' )
#
# sim_dir
# is the directory where the input csv files are located.
#
# random_effect_node_rate_sex =
def read_random_effect_node_rate_sex(sim_dir) :
   assert type(sim_dir) == str
   #
   file_name            = f'{sim_dir}/random_effect.csv'
   random_effect_table  = at_cascade.csv.read_table(file_name)
   result               = dict()
   for row in random_effect_table :
      node_name = row['node_name']
      rate_name = row['rate_name']
      sex       = row['sex']
      if node_name not in result :
         result[node_name] = dict()
      if rate_name not in result[node_name] :
         result[node_name][rate_name] = dict()
      if sex in result[node_name][rate_name] :
         msg  = 'simulate.py: {file_name}\n'
         msg += f'node_name = {node_name},'
         msg += f'rate_name = {rate_name},'
         msg += f'sex = {sex} appears twice'
         sys.exit(msg)
      result[node_name][rate_name][sex] = float( row['random_effect'] )
   return result
#
# ----------------------------------------------------------------------------
# write_random_effect_csv
#
# float_format
# is the format used to write the random effects to random_effects.csv
#
# sim_dir
# is the directory where random_effects.csv is written.
#
# random_effect_node_rate_sex[node_name][rate_name][sex]
# is the random effect corresponding to the specified
# node_name, rate_name, and sex.
#
def write_random_effect_csv(
   float_format, sim_dir, random_effect_node_rate_sex,
) :
   assert type(float_format) == str
   assert type(sim_dir) == str
   assert type(random_effect_node_rate_sex) == dict
   #
   # random_effect_table
   random_effect_table = list()
   for node_name in random_effect_node_rate_sex :
      for rate_name in random_effect_node_rate_sex[node_name] :
         for sex in  random_effect_node_rate_sex[node_name][rate_name] :
            #
            # row
            row                  = { 'node_name' : node_name, 'sex' : sex }
            row['rate_name']     = rate_name
            random_effect = \
               random_effect_node_rate_sex[node_name][rate_name][sex]
            row['random_effect'] = float_format.format(random_effect)
            #
            # random_effect_table
            random_effect_table.append( row )
   #
   # random_effect.csv
   file_name = f'{sim_dir}/random_effect.csv'
   at_cascade.csv.write_table(file_name, random_effect_table)
#
# ----------------------------------------------------------------------------
# sim_random_effect_node_rate_sex
#
# random_effect_node_rate_sex[node_name][rate_name][sex] :
# is the simulated random effect for the corresponding node, sex and rate.
# For a given node_name, sex and rate_name, the sum of
#  random_effect_node_rate_sex[child_node][rate_name][sex]
# is zero where child_node ranges over values in child_list_node[node_name]
# All the random effects for the root node are zero
# (prent_node_dict[root_node_name] == '' )
#
# random_depend_sex:
# is a bool specifying if the random effects depend on sex.
#
# std_random_effects:
# the value std_random_effects[rate]
# is a float specifying the standard deviation of the random effects
# for rate where rate is pini, iota, chi, or rho.
#
# rate_name_list:
# is the list of rate_names the the random effects are simulated for
#
# parent_node_dict;
# is the mapping from a node_name to its parent node_name.
#
# child_list_node:
# for each node_name in parent_node_dict this is a list of
# node that have node_name as its parent node.
#
# random_effect_node_rate_sex =
def sim_random_effect_node_rate_sex (
   random_depend_sex  ,
   std_random_effects ,
   rate_name_list     ,
   parent_node_dict   ,
   child_list_node    ,
) :
   def recursive_call(
      random_effect_node_rate_sex ,
      node_name                   ,
      rate_name                   ,
      sex                         ,
      std_random_effects          ,
      parent_node_dict            ,
      child_list_node             ,
   ) :
      #
      # child_list
      child_list = child_list_node[node_name]
      #
      # n_children
      n_children = len(child_list)
      if n_children == 0 :
         return None
      assert 1 < n_children
      #
      # factor
      factor  = n_children / (n_children - 1)
      assert type(factor) == float
      factor  = math.sqrt( factor )
      #
      # random_effect_dict, sum_random_effect
      random_effect_dict = dict()
      sum_random_effect  = 0.0
      for child_name in child_list :
         std           = std_random_effects[rate_name]
         random_effect = random.gauss(0.0, factor * std)
         sum_random_effect += random_effect
         random_effect_dict[child_name] = random_effect
      #
      # random_effect_dict
      avg_random_effect = sum_random_effect / n_children
      for child_name in child_list :
         random_effect_dict[child_name] -= avg_random_effect
      #
      # random_effect_node_rate_sex
      for child_name in child_list :
         random_effect_node_rate_sex[child_name][rate_name][sex] = \
            random_effect_dict[child_name]
      #
      # random_effect_node_rate_sex
      for child_name in child_list :
         recursive_call(
            random_effect_node_rate_sex ,
            child_name                  ,
            rate_name                   ,
            sex                         ,
            std_random_effects          ,
            parent_node_dict            ,
            child_list_node             ,
         )
      return None
   #
   # root_node_name
   root_node_name = None
   for node_name in parent_node_dict :
      if parent_node_dict[node_name] == '' :
         root_node_name = node_name
   assert root_node_name != None
   #
   # random_effect_node_rate_sex
   random_effect_node_rate_sex  = dict()
   for node_name in parent_node_dict :
      random_effect_node_rate_sex[node_name] = dict()
      for rate_name in rate_name_list :
         random_effect_node_rate_sex[node_name][rate_name] = dict()
   #
   # random_effect_node_rate_sex[root_node_name]
   for rate_name in rate_name_list :
      for sex in [ 'female', 'male' ] :
         random_effect_node_rate_sex[root_node_name][rate_name][sex] = 0.0

   #
   # random_effect_node_rate_sex
   for rate_name in rate_name_list :
      #
      # sex_list
      if random_depend_sex :
         sex_list = [ 'female', 'male' ]
      else :
         sex_list = [ 'female' ]
      #
      # random_effects_node_rate_sex
      for sex in sex_list :
         recursive_call(
            random_effect_node_rate_sex ,
            root_node_name              ,
            rate_name                   ,
            sex                         ,
            std_random_effects          ,
            parent_node_dict            ,
            child_list_node             ,
         )
   #
   # random_effect_node_rate_sex
   if not random_depend_sex :
      for node_name in random_effect_node_rate_sex :
         for rate_name in random_effect_node_rate_sex[node_name]:
            random_effect_node_rate_sex[node_name][rate_name]['male'] = \
               random_effect_node_rate_sex[node_name][rate_name]['female']
   #
   return random_effect_node_rate_sex
# ----------------------------------------------------------------------------
# create_data_sim_table
#
# valid_integrand_name
# a list of the valid integrand names
#
# parent_node_dict
# a dictionary that mapes nodes to parent nodes.
# the root node is in the keys and has an empty parent.
#
# spline_no_effect_rate
# maps rate_name to a spline that evaluates the rate as a function of
# age and time.
#
# random_effect_node_rate_sex
# maps node_name, rate_name, and sex -> the random effect.
#
# root_covariate_ref
# is the covariate reference value for the root node.
#
#
# multiplier_list_rate[rate_name] :
# is the list of rows, in the multiplier_sim table, that have rate_name
# equal to the specified rate_name.
#
# spline_node_sex_cov
# maps node_name, sex, and covariate to a apline that evaluates the
# covarite as a function of age and time.
#
def create_data_sim_table(
   simulate_table,
   valid_integrand_name        ,
   parent_node_dict            ,
   spline_no_effect_rate       ,
   random_effect_node_rate_sex ,
   root_covariate_ref          ,
   multiplier_list_rate        ,
   spline_node_sex_cov         ,
) :
   #
   # float_format
   n_digits = str( global_option_value['float_precision'] )
   float_format = '{0:.' + n_digits + 'g}'
   #
   # node_set
   node_set = set( parent_node_dict.keys() )
   #
   # s_last, s_start
   if global_option_value['trace'] :
      simulate_id = len( simulate_table )
      print( f'Simulation: total id = {simulate_id:,}' )
   s_last  = time.time()
   s_start = s_last
   #
   # data_sim_table
   data_sim_table = list()
   for (simulate_id, sim_row) in enumerate( simulate_table) :
      line_number = simulate_id + 2
      #
      # s_current
      s_current = time.time()
      if s_current - s_last > 30.0 and global_option_value['trace'] :
            #
            # seconds, s_last
            seconds = s_current - s_start
            s_last  = s_current
            print( f'{simulate_id:,} id, {seconds:.0f} sec' )
      #
      # simulate_id
      if True :
         if simulate_id != int( float(sim_row['simulate_id']) ) :
            msg  = f'csv.simulate: Error at line {line_number} '
            msg += f'in simulate.csv\n'
            msg += f'simulate_id = ' + sim_row['simulate_id']
            msg += ' is not equal line number minus two'
            assert False, msg
      #
      # integrand_name
      integrand_name = sim_row['integrand_name']
      if integrand_name not in valid_integrand_name :
         msg  = f'csv.simulate: Error at line {line_number} '
         msg += f' in simulate.csv\n'
         msg += f'integrand_name = ' + integrand_name
         msg += ' is not a valid integrand name'
         assert False, msg
      #
      # node_name
      node_name = sim_row['node_name']
      if node_name not in node_set :
         msg  = f'csv.simulate: Error at line {line_number} '
         msg += f' in simulate.csv\n'
         msg += f'node_name = ' + node_name
         msg += ' is not in node.csv'
         assert False, msg
      #
      # sex
      sex = sim_row['sex']
      if sex not in [ 'female', 'male', 'both' ] :
         msg  = f'csv.simulate: Error at line {line_number} '
         msg += f' in simulate.csv\n'
         msg += f'sex = ' + sex
         msg += ' is not male, feamle, or both'
         assert False, msg
      #
      # age_lower, age_upper, time_lower, time_upper
      age_lower  = float( sim_row['age_lower'] )
      age_upper  = float( sim_row['age_upper'] )
      time_lower = float( sim_row['time_lower'] )
      time_upper = float( sim_row['time_upper'] )
      #
      # age_mid, time_mid
      age_mid  = ( age_lower  + age_upper )  / 2.0
      time_mid = ( time_lower + time_upper ) / 2.0
      #
      covariate_value_dict = dict()
      for cov_name in root_covariate_ref.keys() :
         covariate_value_dict[cov_name] = eval_spline(
            spline_node_sex_cov, node_name, sex, cov_name, age_mid, time_mid
         )
      #
      # rate_fun_dict
      rate_fun_dict = get_rate_fun_dict(
         parent_node_dict            ,
         spline_no_effect_rate       ,
         random_effect_node_rate_sex ,
         root_covariate_ref          ,
         multiplier_list_rate        ,
         node_name                   ,
         sex                         ,
         spline_node_sex_cov         ,
      )
      #
      # data_row
      # Note that changes to data_row will also change covariate_value_dict
      data_row = covariate_value_dict
      data_row['simulate_id'] = simulate_id
      #
      # grid
      integrand_step_size = global_option_value['integrand_step_size']
      grid = average_integrand_grid(
         integrand_step_size, age_lower, age_upper, time_lower, time_upper
      )
      #
      # avg_integrand
      abs_tol = global_option_value['absolute_tolerance']
      avg_integrand = dismod_at.average_integrand(
         rate_fun_dict, integrand_name, grid, abs_tol,
      )
      # numpy uses its own type for floats
      avg_integrand = float( avg_integrand )
      #
      # data_row['meas_mean']
      meas_mean             = avg_integrand
      data_row['meas_mean'] = meas_mean
      #
      # data_row['meas_std']
      meas_std_cv          = float( sim_row['meas_std_cv'] )
      meas_std_min         = float( sim_row['meas_std_min'] )
      meas_std             = max(meas_std_min, meas_std_cv * meas_mean )
      data_row['meas_std'] = meas_std
      #
      # data_row['meas_value']
      meas_value             = random.gauss(meas_mean, meas_std )
      meas_value             = max(meas_value, 0.0)
      data_row['meas_value'] = meas_value
      #
      # data_row
      for key in data_row :
         value = data_row[key]
         if type( value ) == float :
            data_row[key] = float_format.format(value)
      #
      # data_sim_table
      data_sim_table.append( data_row )
   #
   # seconds
   seconds = s_current - s_start
   if global_option_value['trace'] :
      print( f'End simulation: total seconds = {seconds:.0f}' )
   #
   return data_sim_table
# ----------------------------------------------------------------------------
# BEGIN_SIMULATE
# at_cascade.csv.simulate
def simulate(sim_dir) :
   assert type(sim_dir) == str
   # END_SIMULATE
   valid_integrand_name = {
      'Sincidence',
      'remission',
      'mtexcess',
      'mtother',
      'mtwith',
      'susceptible',
      'withC',
      'prevalence',
      'Tincidence',
      'mtspecific',
      'mtall',
      'mtstandard',
      'relrisk',
   }
   #
   # input_table
   input_table = dict()
   file_name   = f'{sim_dir}/option_sim.csv'
   input_table['option_sim'] = at_cascade.csv.read_table(file_name)
   #
   # global_option_value
   set_global_option_value( sim_dir, input_table['option_sim'] )
   #
   # random_seed
   random.seed( int( global_option_value['random_seed'] ) )
   #
   #
   if global_option_value['trace'] :
      print('Reading csv files:')
   input_list  = [
      'node',
      'covariate',
      'multiplier_sim',
      'no_effect_rate',
      'simulate',
   ]
   for name in input_list :
      file_name         = f'{sim_dir}/{name}.csv'
      input_table[name] = at_cascade.csv.read_table(file_name)
      at_cascade.csv.check_table(file_name, input_table[name])
   #
   # input_table['covariate']
   for row in input_table['covariate'] :
      for key in row :
         if key not in { 'node_name', 'sex' } :
            row[key] = float( row[key] )
   #
   if global_option_value['trace'] :
      print('Creating data structures:' )
   #
   # parent_node_dict, child_list_node
   parent_node_dict, child_list_node = \
      get_parent_node_dict( input_table['node'] )
   #
   # spline_node_sex_cov
   node_set = set( parent_node_dict.keys() )
   age_grid, time_grid, spline_node_sex_cov = at_cascade.csv.covariate_spline(
      input_table['covariate'], node_set
   )
   #
   # root_node_name
   root_node_name = None
   for node_name in parent_node_dict :
      if parent_node_dict[node_name] == '' :
         if root_node_name != None :
            msg = 'node.csv has more than one node with no parent'
            assert False, msg
         root_node_name = node_name
   if root_node_name == None :
      msg = 'node.csv every node has a parent node; i.e, no root node'
      assert False, msg
   #
   # root_covariate_ref
   covariate_table    = input_table['covariate']
   covariate_table    = at_cascade.csv.covariate_both(covariate_table)
   root_covariate_ref = at_cascade.csv.covariate_avg(
      covariate_table = covariate_table          ,
      node_name       = root_node_name           ,
      sex             = 'both'                   ,
   )
   absolute_covariates = global_option_value['absolute_covariates']
   if absolute_covariates != None :
      for covariate_name in absolute_covariates.split() :
         root_covariate_ref[covariate_name] = 0.0
   #
   # spline_no_effect_rate
   spline_no_effect_rate = get_spline_no_effect_rate(
      input_table['no_effect_rate']
   )
   #
   # random_effect_node_rate_sex
   random_depend_sex   = global_option_value['random_depend_sex']
   rate_name_list      = spline_no_effect_rate.keys()
   std_random_effects  = {
      'pini' : global_option_value['std_random_effects_pini'] ,
      'iota' : global_option_value['std_random_effects_iota'] ,
      'rho'  : global_option_value['std_random_effects_rho']  ,
      'chi'  : global_option_value['std_random_effects_chi']  ,
   }
   #
   # float_format
   n_digits = str( global_option_value['float_precision'] )
   float_format = '{0:.' + n_digits + 'g}'
   #
   if global_option_value['new_random_effects'] :
      random_effect_node_rate_sex = sim_random_effect_node_rate_sex(
         random_depend_sex  ,
         std_random_effects ,
         rate_name_list     ,
         parent_node_dict   ,
         child_list_node    ,
      )
      #
      # random_effect.csv
      write_random_effect_csv(
         float_format, sim_dir, random_effect_node_rate_sex
      )
   else :
      random_effect_node_rate_sex = read_random_effect_node_rate_sex(sim_dir)
   #
   # multiplier_list_rate
   multiplier_list_rate = get_multiplier_list_rate(
      input_table['multiplier_sim']
   )
   #
   # data_sim_table
   data_sim_table = create_data_sim_table(
      input_table['simulate']     ,
      valid_integrand_name        ,
      parent_node_dict            ,
      spline_no_effect_rate       ,
      random_effect_node_rate_sex ,
      root_covariate_ref          ,
      multiplier_list_rate        ,
      spline_node_sex_cov         ,
   )
   #
   # data.csv
   file_name = f'{sim_dir}/data_sim.csv'
   at_cascade.csv.write_table(file_name, data_sim_table)
   #
   if global_option_value['trace'] :
      print( 'csv.simulate done' )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin job_descendant}

Check if A Job is a Descendant of Another
#########################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

job_table
*********
Is the :ref:`create_job_table@job_table` for this analysis.

ancestor_id
***********
is the :ref:`create_job_table@job_table@job_id` for the ancestor job.

descendant_id
*************
is the job_id for the descendant job.

generation
**********
If the job *descendant* is a descendant of job *ancestor_id* ,
the return value *generation* is the number of generations between them.
If the two job ids are the same, *generation* is zero.
If the job *descendant* is not a descendant of job *ancestor_id* ,
the return value is None.

Node Depth Versus Job Depth
***************************
We use ancestor and descendant node to denote the nodes
corresponding to:

   |  *job_table* [ *ancestor_id*] ['fit_node_id']
   |  *job_table* [ *descendant*] ['fit_node_id']

We use ancestor and descendant reference to denote the split reference
values corresponding to:

   |  *job_table* [ *ancestor_id*] ['split_reference_id']
   |  *job_table* [ *descendant*] ['split_reference_id']

If the ancestor reference is equal to the descendant reference,
or if :ref:`option_all_table@refit_split` is false,
*generation* is equal to the number of nodes between the
ancestor and descendant nodes.
Otherwise, *generation* is one more than the number of nodes between the
ancestor and descendant nodes.
(There can be at most one split between any two nodes.

{xrst_end job_descendant}
'''
# -----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.job_descendant
def job_descendant(job_table, ancestor_id, descendant_id) :
   assert type(job_table)   == list
   assert type(ancestor_id)   == int
   assert type(descendant_id) == int
   # END_DEF
   #
   # generation
   generation = 0
   job_id     = descendant_id
   while job_id != None and job_id != ancestor_id :
      generation += 1
      job_id      = job_table[job_id]['parent_job_id']
   if job_id == None :
      generation = None
   #
   # BEGIN_RETURN
   assert generation == None or type(generation) == int
   return generation
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin csv.covariate_avg}

Compute Covariate Averages for One Node
#######################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

covariate_table
***************
Is a ``list`` of ``dict`` representation of a
:ref:`csv.simulate@Input Files@covariate.csv` file.
All of the columns in this table have been converted to ``float``
except for *node_name* and *sex* which have type ``str`` .
In addition, *sex* equal to ``both`` may have been added; see
:ref:`csv.covariate_both-name` .

node_name
*********
is a str contain the name of the node that we are computing the average for.

covariate_average
*****************
This return is a dict where the keys are the covariates in covariate.csv
and the values are the average of the corresponding covariate.

sex
***
is the sex that this average is for.

{xrst_end csv.covariate_avg}
'''
# BEGIN_DEF
# at_cascade.csv.covariate_avg
def covariate_avg(covariate_table, node_name, sex) :
   assert type(covariate_table) == list
   assert type(covariate_table[0]) == dict
   assert type(node_name) == str
   assert sex in [ 'female', 'male', 'both' ]
   # END_DEF
   #
   # covariate_name_list
   covariate_name_list = list()
   for key in covariate_table[0].keys() :
      if key not in [ 'node_name', 'sex', 'age', 'time', 'omega' ] :
         covariate_name_list.append(key)
   #
   # covariate_sum, count
   covariate_sum = dict()
   for covariate_name in covariate_name_list :
      covariate_sum[covariate_name] = 0.0
   count = 0
   #
   # covariate_sum, count
   line_number = 0;
   for row in covariate_table :
      line_number += 1
      if row['sex'] not in  { 'female', 'male', 'both' } :
         sex  = row['sex']
         msg  = f'covariate.csv at line {line_number}\n'
         msg += 'sex = {sex} is not female or male'
         assert False, msg
      #
      if node_name == row['node_name'] and row['sex'] == sex :
         count += 1
         for covariate_name in covariate_name_list :
            covariate_sum[covariate_name] += float( row[covariate_name] )
   #
   if count == 0 :
      msg  = f'node "{node_name}" does not appear with sex "{sex}" '
      msg += 'covariate_table.'
      assert False, msg
   #
   # covariate_average
   covariate_average = dict()
   for covariate_name in covariate_name_list :
      covariate_average[covariate_name] = \
         covariate_sum[covariate_name] / count
   #
   # BEGIN_RETURN
   # ...
   assert type(covariate_average) == dict
   for value in covariate_average.values() :
      assert type(value) == float
   return covariate_average
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin fit_parallel}
{xrst_spell
  cpus
}

Fit With Specified Maximum Number of Processes
##############################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

job_table
*********
This is a :ref:`create_job_table@job_table` containing all the jobs
necessary to fit the :ref:`glossary@fit_goal_set` .

start_job_id
************
This is the :ref:`create_job_table@job_table@job_id`
for the starting job.
The fit_parallel routine will not return until this job,
and all it descendants in the job table, have been run,
or an error occurs that prevents a job from completing..

all_node_database
*****************
:ref:`fit_one_job@all_node_database`

node_table
**********
:ref:`fit_one_job@node_table`

fit_integrand
*************
:ref:`fit_one_job@fit_integrand`

skip_start_job
**************
If this is true (false) the job corresponding to *start_job_id*
will be skipped (will not be skipped).
If this argument is true, the start job must have already been run.
This is useful when using :ref:`continue_cascade-name` .

max_number_cpu
**************
This is the maximum number of cpus (processes) to use.
This must be greater than zero.
If it is one, the jobs are run sequentially; i.e., not in parallel.

fit_type_list
*************
This is a list with one or two elements
and its possible elements are ``both`` and ``fixed``.
For each job, the first type of fit is attempted.
If it fails, and there is a second type of fit, it is attempted.
If it also fails, the corresponding job fails.

shared_unique
*************
All of these jobs us the following two python multiprocessing
shared memory names:

|  *shared_memory_prefix* _ *job_name* *shared_unique* _number_cpu_in_use
|  *shared_memory_prefix* _ *job_name* *shared_unique* _job_status

#. *job_name* is *job_table* [ *start_job_id* ] [ ``"job_name"`` ]

#. :ref:`option_all_table@shared_memory_prefix` is specified
   in the option all table.

#. *shared_unique* is text that makes this shared memory name unique among
   all the currently running calls to fit_parallel.
   It is suggested that you use the empty string for this value unless you
   are running more than one call with the same prefix and job name.

trace.out
*********
If the *max_number_cpu* is one, standard output is not redirected.
Otherwise, standard output for each job is written to a file called
``trace.out`` in the same directory as the database for the job.

{xrst_end fit_parallel}
'''
# ----------------------------------------------------------------------------
import multiprocessing
import numpy
import at_cascade
import dismod_at
# ----------------------------------------------------------------------------
# shared_memory_prefix = get_shared_memory_prefix(all_node_database)
def get_shared_memory_prefix(all_node_database) :
   assert type(all_node_database) == str
   #
   connection           = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table     = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   shared_memory_prefix = ""
   for row in option_all_table :
      if row['option_name'] == 'shared_memory_prefix' :
         shared_memory_prefix = row['option_value']
   return shared_memory_prefix
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.fit_parallel
def fit_parallel(
   job_table         ,
   start_job_id      ,
   all_node_database ,
   node_table        ,
   fit_integrand     ,
   skip_start_job    ,
   max_number_cpu    ,
   fit_type_list     ,
   shared_unique     ,
) :
   #
   assert type(job_table)         == list
   assert type(start_job_id)      == int
   assert type(all_node_database) == str
   assert type(node_table)        == list
   assert type(fit_integrand)     == set
   assert type(skip_start_job)    == bool
   assert type(max_number_cpu)    == int
   assert type(fit_type_list)     == list
   assert type(shared_unique)     == str
   # END_DEF
   # ----------------------------------------------------------------------
   # job_status_name
   job_status_name = [
      'skip' , # This is a prior only job and completed by the parent fit
      'wait' , # job is waiting for it's parent job to finish
      'ready', # job is readiy to run
      'run'  , # job is running
      'done' , # job finished running
      'error', # job had an exception and could not recover
      'abort', # job is a descendant of a job that could not recover
   ]
   job_status_skip  = job_status_name.index( 'skip' )
   job_status_wait  = job_status_name.index( 'wait' )
   job_status_ready = job_status_name.index( 'ready' )
   job_status_run   = job_status_name.index( 'run' )
   job_status_done  = job_status_name.index( 'done' )
   job_status_error = job_status_name.index( 'error' )
   job_status_abort = job_status_name.index( 'abort' )
   # ----------------------------------------------------------------------
   # shared_memory_prefix_plus
   shared_memory_prefix = get_shared_memory_prefix(all_node_database)
   start_name           = job_table[start_job_id]['job_name']
   shared_memory_prefix_plus = \
      f'{shared_memory_prefix}_{start_name}{shared_unique}'
   print(f'create: {shared_memory_prefix_plus} shared memory')
   # -------------------------------------------------------------------------
   # shared_number_cpu_inuse_name
   shared_number_cpu_inuse_name = \
      shared_memory_prefix_plus + '_number_cpu_inuse'
   #
   # shm_number_cpu_inuse, shared_number_cpu_inuse
   tmp    = numpy.empty(1, dtype = int )
   mapped = at_cascade.map_shared(shared_number_cpu_inuse_name)
   shm_number_cpu_inuse = multiprocessing.shared_memory.SharedMemory(
      create = True, size = tmp.nbytes, name = mapped
   )
   shared_number_cpu_inuse = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_number_cpu_inuse.buf
   )
   # -------------------------------------------------------------------------
   # shared_job_status_name
   shared_job_status_name = shared_memory_prefix_plus + '_job_status'
   #
   # shm_job_status, shared_job_status
   tmp    = numpy.empty(len(job_table), dtype = int )
   mapped = at_cascade.map_shared(shared_job_status_name)
   shm_job_status = multiprocessing.shared_memory.SharedMemory(
      create = True, size = tmp.nbytes, name = mapped
   )
   shared_job_status = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_job_status.buf
   )
   # -------------------------------------------------------------------------
   #
   # shm_list
   shm_list = [
         shm_number_cpu_inuse,
         shm_job_status,
   ]
   #
   # shared_number_cpu_inuse
   shared_number_cpu_inuse[0] = 1
   #
   # shared_job_status
   for job_id in range( len(job_table) ) :
      if job_table[job_id]['prior_only'] :
         shared_job_status[job_id] = job_status_skip
      else :
         shared_job_status[job_id]  = job_status_wait
   if skip_start_job :
      shared_job_status[start_job_id] = job_status_done
      #
      # shared_job_status[child_job_id]
      start_child_job_id    = job_table[start_job_id ]['start_child_job_id']
      end_child_job_id      = job_table[start_job_id ]['end_child_job_id']
      child_range = range(start_child_job_id, end_child_job_id)
      for child_job_id in child_range :
         if not job_table[child_job_id]['prior_only'] :
            shared_job_status[child_job_id] = job_status_ready
   else :
      shared_job_status[start_job_id] = job_status_run
   #
   # master_process
   master_process = True
   #
   # shared_lock
   shared_lock = multiprocessing.Lock()
   #
   # shared_event
   # shared memory has changed
   shared_event = multiprocessing.Event()
   shared_event.set()
   #
   # fit_one_process
   at_cascade.fit_one_process(
      job_table,
      start_job_id,
      all_node_database,
      node_table,
      fit_integrand,
      skip_start_job,
      max_number_cpu,
      master_process,
      fit_type_list,
      job_status_name,
      shared_job_status_name,
      shared_number_cpu_inuse_name,
      shared_lock,
      shared_event,
   )
   #
   # shared_number_cpu_inuse
   if shared_number_cpu_inuse[0] != 1 :
      n_inuse = shared_number_cpu_inuse[0]
      msg =f'{shared_memory_prefix_plus}_number_cpu_inuse[0] = {n_inuse}'
      assert False, msg
   #
   # shared_job_status
   for job_id in range(0, len(job_table) ):
      status = shared_job_status[job_id]
      assert status in \
         [job_status_done, job_status_error, job_status_abort, job_status_skip]
   #
   # free shared memory objects
   print(f'remove: {shared_memory_prefix_plus} shared memory')
   for shm in shm_list :
      shm.close()
      shm.unlink()
   #
   return



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-26 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin module}
{xrst_spell
   cov
}

The at_cascade Python Module
############################

at_cascade.version
******************
The version number for this copy of at_cascade.
{xrst_code py}'''
version = '2026.0.0'
'''{xrst_code}

at_cascade.constant_table_list
******************************
Some of the dismod_at input tables are the same for every fit of a cascade.
The :ref:`fit_or_root_class-name` uses the
:ref:`glossary@root_database` to get the value for these tables:
{xrst_code py} '''
constant_table_list = [
   'age',
   'data',
   'density',
   'integrand',
   'node',
   'rate_eff_cov',
   'subgroup',
   'time',
   'weight',
   'weight_grid',
]
'''{xrst_code}

.. BEGIN_SORT_THIS_LINE_PLUS_2
{xrst_toc_table
   at_cascade/add_log_entry.py
   at_cascade/avgint_parent_grid.py
   at_cascade/bilinear.py
   at_cascade/cascade_root_node.py
   at_cascade/check_cascade_node.py
   at_cascade/check_log.py
   at_cascade/clear_shared.py
   at_cascade/com_cov_reference.py
   at_cascade/continue_cascade.py
   at_cascade/copy_other_tbl.py
   at_cascade/copy_root_db.py
   at_cascade/create_all_node_db.py
   at_cascade/create_job_table.py
   at_cascade/create_shift_db.py
   at_cascade/data_include.py
   at_cascade/empty_avgint_table.py
   at_cascade/empty_directory.py
   at_cascade/extract_avgint.py
   at_cascade/fit_one_job.py
   at_cascade/fit_one_process.py
   at_cascade/fit_or_root_class.py
   at_cascade/fit_parallel.py
   at_cascade/get_cov_info.py
   at_cascade/get_database_dir.py
   at_cascade/get_fit_children.py
   at_cascade/get_fit_integrand.py
   at_cascade/get_freeze_dict.py
   at_cascade/get_parent_node.py
   at_cascade/get_var_id.py
   at_cascade/job_descendant.py
   at_cascade/map_shared.py
   at_cascade/move_table.py
   at_cascade/no_ode_fit.py
   at_cascade/omega_constraint.py
   at_cascade/table_exists.py
   at_cascade/table_name2id.py
}
.. END_SORT_THIS_LINE_MINUS_2

{xrst_end module}
'''

# The file at_cascade/csv/__init__.py should be a sibling (not child)
# of this file in the xrst table of contents.

# BEGIN_SORT_THIS_LINE_PLUS_1
from .                      import csv
from .add_log_entry         import add_log_entry
from .avgint_parent_grid    import avgint_parent_grid
from .bilinear              import bilinear
from .cascade_root_node     import cascade_root_node
from .check_cascade_node    import check_cascade_node
from .check_log             import check_log
from .clear_shared          import clear_shared
from .com_cov_reference     import com_cov_reference
from .continue_cascade      import continue_cascade
from .copy_other_tbl        import copy_other_tbl
from .copy_root_db          import copy_root_db
from .create_all_node_db    import create_all_node_db
from .create_job_table      import create_job_table
from .create_shift_db       import create_shift_db
from .data_include          import data_include
from .empty_avgint_table    import empty_avgint_table
from .empty_directory       import empty_directory
from .extract_avgint        import extract_avgint
from .fit_one_job           import fit_one_job
from .fit_one_process       import fit_one_process
from .fit_or_root_class     import fit_or_root_class
from .fit_parallel          import fit_parallel
from .get_cov_info          import get_cov_info
from .get_database_dir      import get_database_dir
from .get_fit_children      import get_fit_children
from .get_fit_integrand     import get_fit_integrand
from .get_freeze_dict       import get_freeze_dict
from .get_parent_node       import get_parent_node
from .get_var_id            import get_var_id
from .job_descendant        import job_descendant
from .map_shared            import map_shared
from .move_table            import move_table
from .no_ode_fit            import no_ode_fit
from .omega_constraint      import omega_constraint
from .table_exists          import table_exists
from .table_name2id         import table_name2id
# END_SORT_THIS_LINE_MINUS_1



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin avgint_parent_grid}

Predicts Rates and Covariate Multipliers on Parent Grid
#######################################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
Create an avgint table that can be used to predict the rates and covariate
multipliers on the parent grid.
For child nodes, the predictions are for the
:ref:`split_reference_table@split_reference_value` for the parent node.
For the parent node, the predictions are for all the possible
split reference values.

all_node_database
*****************
is a python string containing the name of the :ref:`all_node_db-name`.
This argument can't be ``None``.

fit_database
************
is a python string containing the name of the :ref:`glossary@fit_database`.
A new avgint table will be placed in this database,
the previous avgint table in this database is lost,
and there are no other changes to the database.
This argument can't be ``None``.

job_table
*********
This is a :ref:`create_job_table@job_table` containing the jobs
necessary to fit the :ref:`glossary@fit_goal_set`.
If this is ``None`` , we are doing predictions for the same node and
split reference id a in *fit_database*
(This is only used by :ref:`no_ode_fit-name` .)

fit_job_id
**********
This is the :ref:`create_job_table@job_table@job_id`
for the job fits the fit_database.
This is not used when *job_table* is ``None`` .

parent_node
===========
We use *parent_node* to refer to the parent node in the
dismod_at option table in the fit_database.

avgint Table
************
The new avgint table has all the standard dismod_at columns
plus the extra columns ( beginning with c\_ ) listed below.
This avgint table enables predictions for all the covariate multipliers
and all the rates. Note that the rates (covariate multipliers) depend
(do not depend) on the covariate reference value.

1. If *job_table* is ``None`` , the avgint table enables predictions
   at the same covariate reference values as for the parent_node
   in *fit_database* . Otherwise, see the cases below.

2. This avgint table enables predictions at the covariate reference values
   corresponding to each (node_id, split_reference_id) pair that are
   children of the fit job.

c_age_id
========
This column identifies a row in the age table of the
fit_database that this prediction is for.

c_time_id
=========
This column identifies a row in the time table of the
fit_database that this prediction is for.

c_split_reference_id
====================
This column identifies a row in the split_reference table of the
all_node_database that this prediction is for.
If the split_reference table is empty (non-empty) the value
will be (will not) be ``None``.
It is also ``None`` for the covariate multiplier predictions.

Rectangular Grid
================
For each covariate multiplier that has non-null group smoothing, all of the
age time pairs in the smoothing are represented in the new avgint table.
For each :ref:`child job<glossary@child jobs>` ,
each rate that has non-null parent smoothing,
and each age time pair in the smoothing,
there is a row in the new avgint table.

{xrst_end avgint_parent_grid}
'''
# ----------------------------------------------------------------------------
import copy
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# cov_reference_list =
def get_cov_reference_list(
   n_covariate, cov_reference_table, node_id, split_reference_id
) :
   #
   #
   cov_reference_list = n_covariate * [None]
   for row in cov_reference_table :
      if row['node_id'] == node_id :
         if row['split_reference_id'] == split_reference_id :
            covariate_id = row['covariate_id']
            if covariate_id < len(cov_reference_list) :
               cov_reference_list[covariate_id] = row['reference_value']
   if None in cov_reference_list :
      covariate_id = cov_reference_list.index(None)
      msg  = 'all_node database: cov_reference table: '
      msg += 'No row has the following values:\n'
      msg += f'node_id = {node_id}, '
      msg += f'split_reference_id = {split_reference_id}, '
      msg += f'covariate_id = {covariate_id}'
      assert False, msg
   #
   return cov_reference_list
# ----------------------------------------------------------------------------
# This routine is very similar to get_child_job_table in create_job_table.
# Perhaps there is a good way to combine these two routines.
#
# child_job_list =
def possible_child_job_list(
   # all node database tables
   option_all_table      ,
   node_split_table      ,
   split_reference_table ,
   # fit node database tables
   fit_node_id           ,
   node_table            ,
   covariate_table       ,
) :
   assert type(fit_node_id) == int
   tables = (
      option_all_table      ,
      node_split_table      ,
      split_reference_table ,
      node_table            ,
      covariate_table       ,
   )
   for table in tables :
      assert type(table) == list
      if len(table) > 0 :
         assert type( table[0] == dict )
   #
   # option_all_dict
   option_all_dict = dict()
   for row in option_all_table :
      option_all_dict[ row['option_name'] ] = row['option_value']
   #
   # fit_node_children
   fit_node_children = list()
   for (node_id, row) in enumerate(node_table) :
      if row['parent'] == fit_node_id :
         fit_node_children.append(node_id)
   #
   # refit_split
   if 'refit_split' in option_all_dict :
      refit_split = option_all_dict['refit_split']
      assert refit_split in [ 'true', 'false' ]
      refit_split = refit_split == 'true'
   else :
      refit_split = False
   #
   # root_node_name
   assert 'root_node_name' in option_all_dict
   root_node_name = option_all_dict['root_node_name']
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(node_table, 'node', root_node_name)
   #
   # root_split_reference_id
   if 'root_split_reference_name' in option_all_dict :
      root_split_reference_name = option_all_dict['root_split_reference_name']
      root_split_reference_id   = at_cascade.table_name2id(
         split_reference_table,
         'split_reference',
         root_split_reference_name
      )
   else :
      root_split_reference_id   = None
      if refit_split :
         msg  = 'option_all_table: refit_split is true and '
         msg += ' root_split_reference_name does not appear'
         assert False, msg
   #
   # fit_split_reference_id
   cov_info = at_cascade.get_cov_info(
      option_all_table, covariate_table, split_reference_table
   )
   if 'split_reference_id' not in cov_info :
      fit_split_reference_id = None
   else :
      fit_split_reference_id = cov_info['split_reference_id']
   #
   # Special case where there is no splitting covariate
   if fit_split_reference_id == None :
      child_job_list = list()
      for shift_node_id in fit_node_children :
         child_job = (shift_node_id, None)
         child_job_list.append(child_job)
      return child_job_list
   #
   # already_split
   already_split = root_split_reference_id != fit_split_reference_id
   #
   # node_split_set
   node_split_set = set()
   for row in node_split_table :
      node_split_set.add( row['node_id'] )
   #
   # shift_reference_set
   if already_split or fit_node_id not in node_split_set :
      shift_reference_set = { fit_split_reference_id }
   else :
      shift_reference_set = set( range( len(split_reference_table) ) )
      shift_reference_set.remove( root_split_reference_id )
   #
   #
   # child_job_list
   child_job_list = list()
   if fit_node_id in node_split_set and not already_split and refit_split :
      for shift_split_reference_id in shift_reference_set :
         child_job = (fit_node_id, shift_split_reference_id)
         child_job_list.append(child_job)
   else :
      for shift_split_reference_id in shift_reference_set :
         for shift_node_id in fit_node_children :
            child_job = (shift_node_id, shift_split_reference_id)
            child_job_list.append(child_job)
   #
   return child_job_list
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.avgint_parent_grid
def avgint_parent_grid(
   all_node_database = None ,
   fit_database      = None ,
   job_table         = None ,
   fit_job_id        = None ,
) :
   assert type(all_node_database)  == str
   assert type(fit_database) == str
   assert type(job_table) == list or job_table == None
   assert type(fit_job_id) == int or fit_job_id == None
   # END_DEF
   #
   # option_all_table
   connection = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   get_table             = dismod_at.get_table_dict
   option_all_table      = get_table(connection, 'option_all')
   node_split_table      = get_table(connection, 'node_split')
   split_reference_table = get_table(connection, 'split_reference')
   cov_reference_table   = get_table(connection, 'cov_reference')
   connection.close()
   #
   # root_database
   root_database      = None
   for row in option_all_table :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
   assert root_database != None
   #
   # fit_tables
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   fit_tables = dict()
   for name in [
      'age',
      'covariate',
      'integrand',
      'mulcov',
      'node',
      'option',
      'rate',
      'smooth_grid',
      'time',
   ] :
      fit_tables[name] = fit_or_root.get_table(name)
   fit_or_root.close()
   #
   # split_covariate_id, fit_split_reference_id, fit_split_reference
   split_covariate_id     = None
   fit_split_reference_id = None
   fit_split_reference    = None
   cov_info = at_cascade.get_cov_info(
      option_all_table, fit_tables['covariate'], split_reference_table
   )
   if 'split_covariate_id' in cov_info :
      split_covariate_id     = cov_info['split_covariate_id']
      fit_split_reference_id = cov_info['split_reference_id']
      split_reference_list   = cov_info['split_reference_list']
      fit_split_reference    = split_reference_list[fit_split_reference_id]
   #
   # minimum_age_id
   minimum_age_id = 0
   minimum_age    = fit_tables['age'][minimum_age_id]['age']
   for (age_id, row) in enumerate(fit_tables['age']) :
      if row['age'] < minimum_age :
         minimum_age_id = age_id
         minimum_age    = row['age']
   #
   # n_covariate
   n_covariate = len( fit_tables['covariate'] )
   #
   # parent_node_id
   parent_node_name = None
   for row in fit_tables['option'] :
      assert row['option_name'] != 'parent_node_id'
      if row['option_name'] == 'parent_node_name' :
         parent_node_name = row['option_value']
   assert parent_node_name is not None
   parent_node_id = at_cascade.table_name2id(
      fit_tables['node'], 'node', parent_node_name
   )
   #
   # child_job_list
   child_job_list = possible_child_job_list(
      option_all_table       = option_all_table,
      node_split_table       = node_split_table ,
      split_reference_table  = split_reference_table ,
      fit_node_id            = parent_node_id,
      node_table             = fit_tables['node'],
      covariate_table        = fit_tables['covariate'],
   )
   #
   # cov_reference_dict
   n_covariate  = len( fit_tables['covariate'] )
   cov_reference_dict = dict()
   if job_table == None :
      #
      # cov_reference_list
      cov_reference_list = get_cov_reference_list(
         n_covariate,
         cov_reference_table,
         parent_node_id,
         fit_split_reference_id
      )
      # cov_reference[ (parent_node_id, fit_split_reference_id) ]
      key                     = (parent_node_id, fit_split_reference_id)
      cov_reference_dict[key] = cov_reference_list
   else :
      #
      # (shift_node_id, shift_split_reference_id)
      for child_job in child_job_list :
         (shift_node_id, shift_split_reference_id) = child_job
         #
         # cov_reference_list
         node_id = fit_tables['node'][shift_node_id]['parent']
         assert shift_node_id == parent_node_id or node_id == parent_node_id
         cov_reference_list = get_cov_reference_list(
            n_covariate,
            cov_reference_table,
            shift_node_id,
            shift_split_reference_id
         )
         #
         # cov_reference[ (shift_node_id, shift_split_reference_id) ]
         key                     = child_job
         cov_reference_dict[key] = cov_reference_list
   #
   # tbl_name
   tbl_name = 'avgint'
   #
   # col_name
   col_name = [
      'integrand_id',
      'node_id',
      'subgroup_id',
      'weight_id',
      'age_lower',
      'age_upper',
      'time_lower',
      'time_upper',
   ]
   #
   # col_tyype
   col_type = [
      'integer',
      'integer',
      'integer',
      'integer',
      'real',
      'real',
      'real',
      'real',
   ]
   #
   # add covariates to col_name and col_type
   for covariate_id in range( n_covariate ) :
      col_name.append( 'x_' + str(covariate_id) )
      col_type.append( 'real' )
   #
   # add the smoothing grid columns to col_name and col_type
   col_name += [ 'c_age_id', 'c_time_id', 'c_split_reference_id' ]
   col_type += 3 * ['integer']
   #
   # name_rate2integrand
   name_rate2integrand = {
      'pini':   'prevalence',
      'iota':   'Sincidence',
      'rho':    'remission',
      'chi':    'mtexcess',
   }
   #
   # initialize row_list
   row_list = list()
   #
   # mulcov_id
   for mulcov_id in range( len( fit_tables['mulcov'] ) ) :
      #
      # mulcov_row
      mulcov_row = fit_tables['mulcov'][mulcov_id]
      #
      # group_smooth_id
      group_smooth_id = mulcov_row['group_smooth_id']
      if not group_smooth_id is None :
         #
         # integrand_id
         integrand_name  = 'mulcov_' + str(mulcov_id)
         integrand_id    = at_cascade.table_name2id(
            fit_tables['integrand'], 'integrand', integrand_name
         )
         #
         # grid_row
         for grid_row in fit_tables['smooth_grid'] :
            if grid_row['smooth_id'] == group_smooth_id :
               #
               # age_id
               age_id    = grid_row['age_id']
               age_lower = fit_tables['age'][age_id]['age']
               age_upper = age_lower
               #
               # time_id
               time_id    = grid_row['time_id']
               time_lower = fit_tables['time'][time_id]['time']
               time_upper = time_lower
               #
               # row
               node_id            = None
               subgroup_id        = 0
               weight_id          = None
               split_reference_id = None
               row = [
                  integrand_id,
                  node_id,
                  subgroup_id,
                  weight_id,
                  age_lower,
                  age_upper,
                  time_lower,
                  time_upper,
               ]
               row += n_covariate * [ None ]
               row += [ age_id, time_id, split_reference_id ]
               #
               # add to row_list
               row_list.append( row )
   #
   # rate_name
   for rate_name in name_rate2integrand :
      #
      # rate_id
      rate_id = at_cascade.table_name2id(
         fit_tables['rate'], 'rate', rate_name
      )
      #
      # parent_smooth_id
      parent_smooth_id = fit_tables['rate'][rate_id]['parent_smooth_id']
      if not parent_smooth_id is None :
         #
         # integrand_id
         integrand_name  = name_rate2integrand[rate_name]
         integrand_id    = at_cascade.table_name2id(
            fit_tables['integrand'], 'integrand', integrand_name
         )
         #
         # grid_row
         for grid_row in fit_tables['smooth_grid'] :
            if grid_row['smooth_id'] == parent_smooth_id :
               #
               # age_id
               age_id    = grid_row['age_id']
               age_lower = fit_tables['age'][age_id]['age']
               age_upper = age_lower
               #
               # prior for pini must use age index zero
               if rate_name == 'pini' :
                  assert age_id == minimum_age_id
               #
               # time_id
               time_id    = grid_row['time_id']
               time_lower = fit_tables['time'][time_id]['time']
               time_upper = time_lower
               #
               # key
               for key in cov_reference_dict :
                  #
                  # node_id
                  node_id = key[0]
                  #
                  # split_reference_id
                  split_reference_id = key[1]
                  #
                  # row
                  subgroup_id = 0
                  weight_id   = None
                  row = [
                     integrand_id,
                     node_id,
                     subgroup_id,
                     weight_id,
                     age_lower,
                     age_upper,
                     time_lower,
                     time_upper,
                  ]
                  row += cov_reference_dict[key]
                  row += [ age_id, time_id, split_reference_id ]
                  #
                  # add to row_list
                  row_list.append( row )
   #
   # put new avgint table in fit_database
   connection    = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   command       = 'DROP TABLE IF EXISTS ' + tbl_name
   dismod_at.sql_command(connection, command)
   dismod_at.create_table(connection, tbl_name, col_name, col_type, row_list)
   connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin csv.pre_one_job}
{xrst_spell
  avg
  avgint
  tru
}

Calculate the predictions for One Fit
#####################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

predict_job_name
****************
This string specifies the node and sex corresponding to this prediction.

fit_dir
*******
This string is the directory name where the input and output csv files
are located.

sim_dir
*******
is the directory name where the csv simulation files are located.

pre_database
************
The directory where *pre_database* is located identifies the
node and sex for this prediction.
The file *pre_database* is a copy of the fit database used for this prediction.
The node and sex for the fit is either the same as for the prediction,
or an ancestor of the node and ex for the prediction.

predict_node_id
***************
This int is the node_id in the node we are predicting for.

all_node_database
*****************
This string is the all node database for this fit.

all_covariate_table
*******************
The list of dict is the in memory representation of the
covariate.csv file

float_precision
***************
This is the number of decimal digits of precision to include for
float values in fit_predict.csv, sam_predict.csv, and tru_predict.csv;
see below.

fit_same_as_predict
*******************
If true, the node and sex for this prediction is the same
as the node and sex for the fit.
Otherwise it is a fit for the closest ancestor that has a
successful cascade fit.
To be specific, ``sample: OK`` is in its
:ref:`fit_one_job@fit_database@log`.


option_predict
**************
This is an in memory representation of
:ref:`csv.predict@Input Files@option_predict.csv` .

Csv Output Files
****************
#. The csv output files are located in the prediction directory; i.e.,
   the directory corresponding to the predictions for this location, sex.
#. The predictions are on the same age, time grid as the covariate file.
#. If *fit_same_as_predict* is true, the following files are created
   (the tru_posterior.csv file is not created when *sim_dir* is None):

   .. csv-table::

      fit_posterior.csv, uses optimal posterior variable values
      sam_posterior.csv, uses samples from the posterior
      tru_posterior.csv, uses simulation variable values for this location,sex

#. If *fit_same_as_predict* is false, the following files are created
   (the tru_prior.csv file is not created when *sim_dir* is None):

   .. csv-table::

      fit_prior.csv,     uses optimal prior variable values
      sam_prior.csv,     uses samples from the prior
      tru_prior.csv,     uses simulation variable values for an ancestor


#. The following columns are included in these files
   (the sample_index column is only included in the sample files):

   .. csv-table::
      :header-rows: 1

      Column,        Meaning
      avgint_id,     This index the value we are predicting
      sample_index,  This index the random samples for each value
      avg_integrand, This is the model value for the prediction
      age_lower,     Lower age limit for averaging this integrand
      age_upper,     Upper age limit for averaging (must equal lower).
      time_lower,    Lower time limit for averaging this integrand
      time_upper,    Upper time limit for averaging (must equal lower).
      node_id,       Identifies the node for this prediction.
      x_j,           Value of the j-th covariate

{xrst_end csv.pre_one_job}
r'''
import dismod_at
import at_cascade
import copy
#
# ----------------------------------------------------------------------------
# Create Diagonsitcs for One Fit
# ##############################
#
# predict_job_name
# ****************
# This string specifies the node and sex corresponding to this fit.
#
# fit_dir
# *******
# This string is the directory name where the input and output csv files
# are located.
#
# pre_database
# ************
# This string is the location, relative to fit_dir, of the dismod_at
# database for a fit.
# It must be located in the pediction directory.
#
# fit_same_as_predict
# *******************
# if true, the the fit location,sex is same as prediction location,sex.
#
# db2csv
# ******
# If fit_same_as_predict and db2csv,
# the `db2csv_command`_ is used to generate the csv files
# corresponding to the *pre_database* .
#
# .. _db2csv_command: https://dismod-at.readthedocs.io/latest/db2csv_command.html
#
# plot
# ****
# If fit_same_as_predict and plot,
# ``data_plot.pdf`` and ``rate_plot.pdf`` corresponding
# to *pre_database* are generated, in the same directory as the
# *pre_database* ; i.e., the prediction directory.
#
import dismod_at
import at_cascade
import copy

def diagnose_one(
   predict_job_name       ,
   fit_dir                ,
   pre_database           ,
   fit_same_as_predict    ,
   db2csv                 ,
   plot                   ,
) :
   assert type(predict_job_name) == str
   assert type(fit_dir) == str
   assert type(pre_database) == str
   assert type( fit_same_as_predict ) == bool
   assert type( db2csv ) == bool
   assert type( plot ) == bool
   #
   # predict_integrand_table
   predict_integrand_table = at_cascade.csv.read_table(
         f'{fit_dir}/predict_integrand.csv'
   )
   #
   # predict_node_dir
   index            = pre_database.rfind('/')
   predict_node_dir = pre_database[: index]
   #
   if fit_same_as_predict and db2csv :
      #
      # db2csv output files
      dismod_at.db2csv_command( pre_database, )
   if fit_same_as_predict and plot :
      #
      # data_plot.pdf
      pdf_file       = f'{predict_node_dir}/data_plot.pdf'
      integrand_list = list()
      for row in predict_integrand_table :
         integrand_name = row['integrand_name']
         if not integrand_name.startswith('mulcov_') :
            integrand_list.append( integrand_name )
      dismod_at.plot_data_fit(
         database       = pre_database           ,
         pdf_file       = pdf_file               ,
         plot_title     = predict_job_name       ,
         max_plot       = 1000                   ,
         integrand_list = integrand_list         ,
      )
      #
      # rate_plot.pdf
      pdf_file = f'{predict_node_dir}/rate_plot.pdf'
      rate_set = { 'pini', 'iota', 'chi', 'omega' }
      dismod_at.plot_rate_fit(
         database       = pre_database           ,
         pdf_file       = pdf_file               ,
         plot_title     = predict_job_name       ,
         rate_set       = rate_set               ,
      )
# ----------------------------------------------------------------------------

# BEGIN_DEF
# at_cascade.csv.pre_one_job
def pre_one_job(
   predict_job_name      ,
   fit_dir               ,
   sim_dir               ,
   pre_database          ,
   predict_node_id       ,
   predict_sex_id        ,
   all_node_database     ,
   all_covariate_table   ,
   float_precision       ,
   fit_same_as_predict   ,
   option_predict        ,
) :
   assert type(predict_job_name) == str
   assert type(fit_dir) == str
   assert sim_dir == None or type(sim_dir) == str
   assert type(pre_database) == str
   assert type(predict_node_id) == int
   assert type(all_node_database) == str
   assert type(all_covariate_table) == list
   assert type( all_covariate_table[0] ) == dict
   assert type( float_precision ) == int
   assert type( fit_same_as_predict ) == bool
   assert type( option_predict ) == dict
   # END_DEF
   #
   #
   # descendant_std_factor, number_sample_predict, zero_meas_value, db2csv, plot
   descendant_std_factor = option_predict['descendant_std_factor']
   number_sample_predict = option_predict['number_sample_predict']
   zero_meas_value       = option_predict['zero_meas_value']
   db2csv                = option_predict['db2csv']
   plot                  = option_predict['plot']
   assert type( db2csv ) == bool
   assert type( plot ) == bool
   assert type( zero_meas_value) == bool
   assert type( number_sample_predict ) == int
   assert type( descendant_std_factor ) == float
   #
   # option_all_table
   connection       = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   #
   # root_database and sample_method
   root_database      = None
   sample_method      = None
   for row in option_all_table :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
      if row['option_name'] == 'sample_method' :
         sample_method      = row['option_value']
   assert root_database != None
   assert sample_method != None
   #
   # fit_covariate_table, integrand_table, node_table
   fit_or_root = at_cascade.fit_or_root_class(
      pre_database, root_database
   )
   fit_covariate_table      = fit_or_root.get_table('covariate')
   integrand_table          = fit_or_root.get_table('integrand')
   node_table               = fit_or_root.get_table('node')
   fit_or_root.close()
   #
   # pre_database
   # add the truth_var table to this database
   if type(sim_dir) == str :
      at_cascade.csv.set_truth(
         sim_dir, pre_database, root_database
      )
   #
   # integrand_id_list
   predict_integrand_table = at_cascade.csv.read_table(
         f'{fit_dir}/predict_integrand.csv'
   )
   integrand_id_list = list()
   for row in predict_integrand_table :
      integrand_id = at_cascade.table_name2id(
         integrand_table, 'integrand', row['integrand_name']
      )
      integrand_id_list.append( integrand_id )
   #
   # predict_node_name
   predict_node_name = node_table[predict_node_id]['node_name']
   #
   # split_reference_table
   split_reference_table = at_cascade.csv.split_reference_table
   #
   # predict_sex_value, predict_sex_name
   row               = split_reference_table[predict_sex_id]
   predict_sex_value = row['split_reference_value']
   predict_sex_name  = row['split_reference_name']
   #
   # avgint_table
   avgint_table = list()
   #
   # male_index_dict
   male_index_dict = dict()
   if predict_sex_name == 'both' :
      for (i_row, row) in enumerate(all_covariate_table) :
         if row['node_name'] == predict_node_name :
            sex  = row['sex']
            age  = float( row['age'] )
            time = float( row['time'] )
            if sex == 'male' :
               if age not in male_index_dict :
                  male_index_dict[age] = dict()
               if time not in male_index_dict[age] :
                  male_index_dict[age][time] = i_row
   #
   # cov_row
   for cov_row in all_covariate_table :
      #
      # select
      if cov_row['sex'] == predict_sex_name :
         select = True
      elif cov_row['sex'] == 'female' and predict_sex_name == 'both' :
         select = True
      else :
         select = False
      select = select and cov_row['node_name'] == predict_node_name
      if select :
         #
         # avgint_row
         age  = float( cov_row['age'] )
         time = float( cov_row['time'] )
         avgint_row = {
            'node_id'         : predict_node_id,
            'subgroup_id'     : 0,
            'weight_id'       : None,
            'age_lower'       : age,
            'age_upper'       : age,
            'time_lower'      : time,
            'time_upper'      : time,
         }
         #
         # covariate_id
         for covariate_id in range( len(fit_covariate_table) ) :
            #
            # covariate_name
            covariate_name = \
               fit_covariate_table[covariate_id]['covariate_name']
            #
            # covariate_value
            if covariate_name == 'one' :
               covariate_value = 1.0
            elif covariate_name == 'sex' :
               covariate_value = predict_sex_value
            else :
               covariate_value = float( cov_row[covariate_name] )
               if predict_sex_name == 'both' :
                  male_row  = all_covariate_table[ male_index_dict[age][time] ]
                  assert male_row['sex'] == 'male'
                  assert cov_row['sex']  == 'female'
                  covariate_value += float( male_row[covariate_name] )
                  covariate_value /= 2.0
            #
            # avgint_row
            key = f'x_{covariate_id}'
            avgint_row[key] = covariate_value
         #
         # integrand_id
         for integrand_id in integrand_id_list :
            avgint_row['integrand_id'] = integrand_id
            avgint_table.append( copy.copy( avgint_row ) )
   #
   # connection
   connection = dismod_at.create_connection(
      pre_database, new = False, readonly = False
   )
   #
   # avgint table
   dismod_at.replace_table(connection, 'avgint', avgint_table)
   #
   # sample_command, sample_ok
   log_table        = dismod_at.get_table_dict(connection, 'log')
   sample_message   = None
   sample_ok        = False
   for row in log_table:
      message_type = row['message_type']
      message      = row['message']
      if message_type == 'command' and message.startswith('end sample') :
         sample_message  = message
         sample_ok       = True
      if message_type == 'command' and message.startswith('begin sample') :
         sample_ok  = False
   if not sample_ok :
      message  = f'pre_one_job: {pre_database}: log table\n'
      message += 'The last begin sample does not have an end sample after it.'
      assert False, message
   #
   # pre_database
   # ensure that sample table has number_sample_predict samples per variable.
   sample_message = sample_message.split()
   number_sample  = int( sample_message[-1] )
   if number_sample != number_sample_predict :
      assert sample_message[0] == 'end'
      assert sample_message[1] == 'sample'
      sample_methopd = sample_message[2]
      variables      = sample_message[3]
      number_sample  = str( number_sample_predict )
      command = [
         'dismod_at',
         pre_database,
         'sample',
         sample_method,
         variables,
         number_sample
      ]
      dismod_at.system_command_prc(command)
   #
   # prefix_list
   prefix_list = list()
   if at_cascade.table_exists(connection, 'fit_var') :
      prefix_list.append( 'fit' )
   if at_cascade.table_exists(connection, 'sample') :
      prefix_list.append( 'sam' )
   if sim_dir != None :
      prefix_list.append( 'tru' )
   connection.close()
   #
   # suffix
   if fit_same_as_predict :
      suffix = 'posterior'
   else :
      suffix = 'prior'
   #
   # predict_node_dir
   index            = pre_database.rfind('/')
   predict_node_dir = pre_database[: index]
   #
   # float_format
   n_digits = str( float_precision )
   float_format = '{0:.' + n_digits + 'g}'
   #
   # prefix
   for prefix in prefix_list :
      #
      # command
      command = [ 'dismod_at', pre_database, 'predict' ]
      if prefix == 'fit' :
         command.append( 'fit_var' )
      elif prefix == 'tru' :
         command.append( 'truth_var' )
      else :
         assert prefix == 'sam'
         command.append( 'sample' )
         if not fit_same_as_predict :
            command += [ 'fit_var', str(descendant_std_factor) ]
      if zero_meas_value :
         command.append( 'zero_meas_value' )
      dismod_at.system_command_prc(command, print_command = False )
      #
      # predict_table, covariate_table
      connection    = dismod_at.create_connection(
            pre_database, new = False, readonly = True
      )
      predict_table   = dismod_at.get_table_dict(connection, 'predict')
      covariate_table = dismod_at.get_table_dict(connection, 'covariate')
      connection.close()
      #
      # sex_covariate_id
      sex_covariate_id = None
      for (covariate_id, row) in enumerate( covariate_table ) :
         if row['covariate_name'] == 'sex' :
            sex_covariate_id = covariate_id
      #
      # predict_table
      for pred_row in predict_table :
         #
         # avgint_row
         avgint_id  = pred_row['avgint_id']
         avgint_row = avgint_table[avgint_id]
         assert avgint_row['node_id'] == predict_node_id
         assert avgint_row[ f'x_{sex_covariate_id}' ] == predict_sex_value
         #
         # predict_table
         for key in avgint_row.keys() :
            pred_row[key] = avgint_row[key]
         avg_integrand             = pred_row['avg_integrand']
         pred_row['avg_integrand'] = float_format.format(avg_integrand)
         if prefix in [ 'fit', 'tru' ] :
            assert pred_row['sample_index'] == None
            del pred_row['sample_index']
         #
      #
      # prefix_suffix.csv
      file_name    = f'{predict_node_dir}/{prefix}_{suffix}.csv'
      at_cascade.csv.write_table(file_name, predict_table)
   #
   diagnose_one(
      predict_job_name     = predict_job_name ,
      fit_dir              = fit_dir ,
      pre_database         = pre_database ,
      fit_same_as_predict  = fit_same_as_predict ,
      db2csv               = db2csv ,
      plot                 = plot ,
   )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin create_shift_db}
{xrst_spell
  dage
  dtime
  var
  std
  delim
}

Create Database With Shifted Covariate References
#################################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Problem
*******
The value and difference priors are treated as independent.
Currently the standard deviations for the difference priors are fixed
at their value for the root level.
In addition, the standard deviations for the value priors are set
using the posterior samples predicted by the parent job fit.
The theory section below discusses the problem with this.
The following options provide a way to avoid this problem:
:ref:`csv.fit@Input Files@option_fit.csv@child_prior_dage` ,
:ref:`csv.fit@Input Files@option_fit.csv@child_prior_dtime` ,
:ref:`option_all_table@shift_prior_dage` ,
:ref:`option_all_table@shift_prior_dtime` .

Theory
******
For this discussion we start our indexing at one
(recall that python starts its indexing at zero).
Suppose that there is only on rate; e.g. iota,
no data, and all the priors are Gaussian.
We use the following notation for the rate's prior:
(The :ref:`fit_info_theory-name` page has a more general discussion
of the analysis below.)

.. csv-table::
   :header-rows: 1
   :delim: ;

   Notation; Meaning
   :math:`r_{i,j}` ; value of the rate at the i-th age and j-th time
   :math:`N`       ; number of age points in the rate grid
   :math:`M`       ; number of time points in the rate grid
   :math:`m_v`     ; mean for the rate values
   :math:`m_a`     ; mean for the rate age differences
   :math:`m_t`     ; mean for the rate time values
   :math:`s_v`     ; standard deviation (std) of the value prior for the rate
   :math:`s_a`     ; std of the age difference prior for the rate
   :math:`s_t`     ; std of the time difference prior for the rate
   :math:`L(r)`    ; the negative log-likelihood

Note that the means and standard deviations actually depend on
the age and time indices (i,j).
We have dropped these indices because they can be inferred from the (i,j)
indices of :math:`r` .

.. math::

   2 L(r) = &
      + \sum_{i,j}
      \left( \frac{ r_{i,j} - m_v } { s_v } \right)^2
      - \log ( 2 \pi s_v^2 )
   \\ &
      + \sum_{i < N , j}
      \left( \frac{ r_{i+1,j} - r_{i,j} - m_a } { s_a } \right)^2
      - \log ( 2 \pi s_a^2 )
   \\ &
      + \sum_{i, j < M}
      \left( \frac{ r_{i,j+1} - r_{i,j} - m_t } { s_t } \right)^2
      - \log ( 2 \pi s_t^2 )

The partial of :math:`L(r)` with respect to :math:`r_{N,M}` is

.. math::

   \frac{ \partial L}{\partial r_{N,M}}
   =
   \frac{ r_{N,M} - m_v } { s_v^2 }
   +
   \frac{ r_{N,M} - r_{N-1,M} - m_a } { s_a^2 }
   +
   \frac{ r_{N,M} - r_{N,M-1} - m_t } { s_t^2 }

The second partial of :math:`L(r)` with respect to :math:`r_{N,M}` is

.. math::

   \frac{ \partial^2 L}{\partial r_{N,M} \partial r_{N,M} }
   =
   s_v^{-2} + s_a^{-2} + s_t^{-2}

The inverse of the Hessian is the covariance matrix for the
estimate of :math:`r` .
Suppose that we want the standard deviation for :math:`r_{N,M}` to be
:math:`\sigma` .
If we approximation the diagonal of the inverse buy the inverse
of the diagonal, it follows that

.. math::

   \sigma^2    & = \left( s_v^{-2} + s_a^{-2} + s_t^{-2} \right)^{-1}
   \\
   \sigma^{-2} & =  s_v^{-2} + s_a^{-2} + s_t^{-2}


all_node_database
*****************
is a python string containing the name of the :ref:`all_node_db-name`.

fit_database
************
is a python string containing the name of a dismod_at database.
This is a :ref:`glossary@fit_database` which
has two predict tables (mentioned below).
These tables are used to create priors in the child node databases.
This argument can't be ``None``.

fit_node
========
We use *fit_node* to refer to the parent node in the
dismod_at option table in the *fit_database*.

sample Table
============
This table is not used if *no_ode_fit* is true.
If *no_ode_fit* is false,
it  contains the results of a dismod_at sample command
for both the fixed and random effects.

c_shift_avgint Table
====================
This is the :ref:`avgint_parent_grid-name` table corresponding
to this fit_database.

c_shift_predict_sample Table
============================
This table is not used if *no_ode_fit* is true.
If *no_ode_fit* is false, it contains the predict table corresponding to a
predict sample command using the c_shift_avgint table.
Note that the predict_id column name was changed to c_shift_predict_sample_id
(which is not the same as sample_id).

c_shift_predict_fit_var Table
=============================
This table contains the predict table corresponding to a
predict fit_var command using the c_shift_avgint table.
Note that the predict_id column name was changed to c_shift_predict_fit_var_id
(which is not the same as var_id).

shift_databases
***************
We use the notation *shift_name* for the keys in this ``dict``.

shift_name
==========
For each *shift_name*, *shift_databases[shift_name]* is the name of
a :ref:`glossary@input_node_database` that is created by this command.
The corresponding directory is assumed to already exist.

split_reference_name
====================
If *shift_name* is a :ref:`split_reference_table@split_reference_name`,
the node corresponding to this shift database is the fit_node.

Child Node
==========
If *shift_name* is the node name for a child of fit_node,
the child is the node corresponding to this shift database.

Fit Node
========
If *shift_name* is the name of the *fit_node*,
the node corresponding to this shift database is the fit_node.
This case is used by :ref:`no_ode_fit-name` to create priors without shifting
the covariate references.

Value Priors
============
If the upper and lower limits are equal,
the value priors in fit_database and the shift_databases
are effectively the same.
Otherwise the mean in the value priors
are replaced using the corresponding values in the
predict tables in the *fit_database*.
If *no_ode_fit* is true (false),
the standard deviations in the value priors are not replaced (are replaced).
Note that if the value prior is uniform,
the standard deviation is not used and the mean is only used to
initialize the optimization.

dage and dtime Priors
=====================
The mean of the dage and dtime priors
are replaced using the corresponding difference in the
predict tables in the *fit_database*.

Log Table
=========
There is no log table in the shifted databases.

no_ode_fit
**********
If this argument is true (false) if the *fit_database*
is (is not) the result of a :ref:`no_ode_fit-name` .
If *no_ode_fit* is false,
the sample table and the c_shift_predict_sample
table must be in the *fit_database*.
In this case both the means and standard deviations in the value priors
are replaced using the results of the fit.
Otherwise, only the means are replaced.

job_table
*********
If *no_ode_fit* is true this argument must be None.
Otherwise it is the :ref:`create_job_table@job_table` for this cascade.

{xrst_end create_shift_db}
'''
# ----------------------------------------------------------------------------
import os
import math
import copy
import shutil
import numpy
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# cov_reference_list =
def get_cov_reference_list(
   n_covariate, cov_reference_table, node_id, split_reference_id
) :
   #
   #
   cov_reference_list = n_covariate * [None]
   for row in cov_reference_table :
      if row['node_id'] == node_id :
         if row['split_reference_id'] == split_reference_id :
            covariate_id = row['covariate_id']
            if covariate_id < len(cov_reference_list) :
               cov_reference_list[covariate_id] = row['reference_value']
   if None in cov_reference_list :
      covariate_id = cov_reference_list.index(None)
      msg  = 'all_node database: cov_reference table: '
      msg += 'No row has the following values:\n'
      msg += f'node_id = {node_id}, '
      msg += f'split_reference_id = {split_reference_id}, '
      msg += f'covariate_id = {covariate_id}'
      assert False, msg
   #
   return cov_reference_list
# ----------------------------------------------------------------------------
def add_index_to_name(table, name_col) :
   row   = table[-1]
   name  = row[name_col]
   ch    = name[-1]
   while name != '' and name[-1] in '0123456789' :
      name = name[: -1]
   if name[-1] == '_' :
      name = name[: -1]
   row[name_col] = name + '_' + str( len(table) )
# ---------------------------------------------------------------------------
def get_age_id_next_list(smooth_table, smooth_grid_table, age_table ) :
   #
   # age_id_set
   age_id_set  = dict()
   for row in smooth_grid_table :
      smooth_id = row['smooth_id']
      age_id    = row['age_id']
      time_id   = row['time_id']
      if smooth_id not in age_id_set :
         age_id_set[smooth_id]  = set()
      age_id_set[smooth_id].add(age_id)
   #
   # age_id_key, time_id_key
   age_id_key  = lambda age_id :  age_table[age_id]['age']
   #
   # age_id_next_list
   age_id_next_list  = list()
   for smooth_id in range( len(smooth_table) ) :
      n_age        = smooth_table[smooth_id]['n_age']
      age_id_list  = sorted(age_id_set[smooth_id], key = age_id_key )
      assert n_age == len(age_id_list)
      age_id_dict  = dict()
      for i in range(n_age) :
         age_id  = age_id_list[i]
         if i + 1 < n_age :
            next_age_id = age_id_list[i+1]
         else :
            next_age_id = None
         age_id_dict[age_id] = next_age_id
      age_id_next_list.append( age_id_dict )
   return age_id_next_list
# ---------------------------------------------------------------------------
def get_time_id_next_list(smooth_table, smooth_grid_table, time_table ) :
   #
   # time_id_set
   time_id_set  = dict()
   for row in smooth_grid_table :
      smooth_id = row['smooth_id']
      time_id    = row['time_id']
      time_id   = row['time_id']
      if smooth_id not in time_id_set :
         time_id_set[smooth_id]  = set()
      time_id_set[smooth_id].add(time_id)
   #
   # time_id_key, time_id_key
   time_id_key  = lambda time_id :  time_table[time_id]['time']
   #
   # time_id_next_list
   time_id_next_list  = list()
   for smooth_id in range( len(smooth_table) ) :
      n_time        = smooth_table[smooth_id]['n_time']
      time_id_list  = sorted(time_id_set[smooth_id], key = time_id_key )
      assert n_time == len(time_id_list)
      time_id_dict  = dict()
      for i in range(n_time) :
         time_id  = time_id_list[i]
         if i + 1 < n_time :
            next_time_id = time_id_list[i+1]
         else :
            next_time_id = None
         time_id_dict[time_id] = next_time_id
      time_id_next_list.append( time_id_dict )
   return time_id_next_list
# ----------------------------------------------------------------------------
# The smoothing for the new shift_table['smooth_grid'] row is the most
# recent smoothing added to shift_table['smooth']; i.e., its smoothing_id
# is len( shift_table['smooth'] ) - 1.
def add_shift_grid_row(
   fit_fit_var,
   fit_sample,
   fit_table,
   shift_table,
   fit_grid_row,
   integrand_id,
   shift_node_id,
   shift_split_reference_id,
   shift_prior_std_factor,
   shift_prior_dage,
   shift_prior_dtime,
   freeze,
   copy_row,
   age_id_next,
   time_id_next,
) :
   assert freeze in { 'prior', 'mean', 'no' }
   if freeze == 'prior' :
      assert copy_row == True
   # -----------------------------------------------------------------------
   # value_prior
   # -----------------------------------------------------------------------
   #
   # split_id
   split_id = shift_split_reference_id
   #
   # fit_prior_id
   fit_prior_id    = fit_grid_row['value_prior_id']
   #
   # shift_const_value
   # shift_value_prior_id
   shift_const_value     = fit_grid_row['const_value']
   shift_value_prior_id  = None
   dage_fit_var          = None
   dtime_fit_var         = None
   if shift_const_value is None :
      #
      # fit_prior_row
      fit_prior_row = fit_table['prior'][fit_prior_id]
      #
      # key
      age_id    = fit_grid_row['age_id']
      time_id   = fit_grid_row['time_id']
      key       = (integrand_id, shift_node_id, split_id, age_id, time_id)
      #
      # lower, upper
      if freeze == 'mean' :
         lower = fit_fit_var[key]
         upper = fit_fit_var[key]
      else :
         lower = fit_prior_row['lower']
         upper = fit_prior_row['upper']
      if lower is None :
         lower = - math.inf
      if upper is None :
         upper = + math.inf
      #
      # shift_const_value, shift_value_prior_id, shift_table['prior']
      if lower == upper :
         shift_const_value  = lower
         assert shift_value_prior_id is None
      else :
         assert shift_const_value is None
         #
         # shift_value_prior_id, shift_prior_row
         shift_value_prior_id  = len( shift_table['prior'] )
         shift_prior_row = copy.copy( fit_prior_row )
         if not copy_row :
            #
            # fit_var, dage_fit_var, dtime_fit_var
            fit_var = fit_fit_var[key]
            if age_id_next[age_id] != None :
               next_age_id = age_id_next[age_id]
               key = (
                  integrand_id, shift_node_id, split_id, next_age_id, time_id
               )
               dage_fit_var = fit_fit_var[key] - fit_var
            if time_id_next[time_id] != None :
               next_time_id = time_id_next[time_id]
               key = (
                  integrand_id, shift_node_id, split_id, age_id, next_time_id
               )
               dtime_fit_var = fit_fit_var[key] - fit_var
            #
            # shift_prior_row['mean']
            mean                     = fit_var
            mean                     = min(mean, upper)
            mean                     = max(mean, lower)
            shift_prior_row['mean']  = mean
            #
            # if no_ode_fit then len(fit_sample) is zero
            if len(fit_sample) > 0 :
               #
               # std
               eta        = fit_prior_row['eta']
               if eta is None :
                  std  = numpy.std(fit_sample[key], mean=mean)
               else:
                  # There is a log transformation of this variable before
                  # passing it to cppad_mixed. Hence its value are gaussian
                  # in log space.
                  #
                  # log_sample
                  log_sample = list()
                  for sample in fit_sample[key] :
                     sample = max( - eta / 5.0 , sample )
                     log_sample.append( math.log( sample + eta ) )
                  #
                  # log_std
                  log_mean = math.log(mean + eta)
                  log_std  = numpy.std(log_sample, mean = log_mean, ddof = 0)
                  #
                  # inverse log transformation
                  std      = (math.exp(log_std) - 1) * (mean + eta)
               #
               # shift_prior_row['std']
               shift_prior_row['std']         = shift_prior_std_factor * std
         #
         # shift_table['prior']
         shift_table['prior'].append( shift_prior_row )
         add_index_to_name( shift_table['prior'], 'prior_name' )
   # -----------------------------------------------------------------------
   # dage_prior
   # -----------------------------------------------------------------------
   if not shift_prior_dage :
      shift_dage_prior_id = None
   else :
      fit_prior_id       = fit_grid_row['dage_prior_id']
      if fit_prior_id == None :
         shift_dage_prior_id = None
      else :
         fit_prior_row      = fit_table['prior'][fit_prior_id]
         shift_prior_row    = copy.copy( fit_prior_row )
         if dage_fit_var is not None :
            mean = dage_fit_var
            if shift_prior_row['lower'] != None :
               if mean < shift_prior_row['lower']  :
                  mean = shift_prior_row['lower']
            if shift_prior_row['upper'] != None :
               if shift_prior_row['lower'] < mean :
                  mean = shift_prior_row['upper']
            shift_prior_row['mean'] = mean
         shift_dage_prior_id  = len( shift_table['prior'] )
         shift_table['prior'].append( shift_prior_row )
         add_index_to_name( shift_table['prior'], 'prior_name' )
   # -----------------------------------------------------------------------
   # dtime_prior
   # -----------------------------------------------------------------------
   if not shift_prior_dtime :
      shift_dtime_prior_id = None
   else :
      fit_prior_id       = fit_grid_row['dtime_prior_id']
      if fit_prior_id == None :
         shift_dtime_prior_id = None
      else :
         fit_prior_row       = fit_table['prior'][fit_prior_id]
         shift_prior_row       = copy.copy( fit_prior_row )
         shift_dtime_prior_id  = len( shift_table['prior'] )
         if dtime_fit_var is not None :
            mean = dtime_fit_var
            if shift_prior_row['lower'] != None :
               if mean < shift_prior_row['lower']  :
                  mean = shift_prior_row['lower']
            if shift_prior_row['upper'] != None :
               if shift_prior_row['lower'] < mean :
                  mean = shift_prior_row['upper']
            shift_prior_row['mean'] = mean
         shift_table['prior'].append( shift_prior_row )
         add_index_to_name( shift_table['prior'], 'prior_name' )
   # -----------------------------------------------------------------------
   # shift_grid_row
   shift_grid_row = copy.copy( fit_grid_row )
   shift_grid_row['value_prior_id']  = shift_value_prior_id
   shift_grid_row['const_value']     = shift_const_value
   shift_grid_row['dage_prior_id']   = shift_dage_prior_id
   shift_grid_row['dtime_prior_id']  = shift_dtime_prior_id
   #
   # shift_table['smooth_grid']
   shift_grid_row['smooth_id']  = len( shift_table['smooth'] ) - 1
   shift_table['smooth_grid'].append( shift_grid_row )
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.create_shift_db
def create_shift_db(
   all_node_database    ,
   fit_database         ,
   shift_databases      ,
   no_ode_fit           = False,
   job_table            = None,
) :
   assert type(all_node_database) == str
   assert type(fit_database) == str
   assert type(shift_databases) == dict
   assert type(no_ode_fit) == bool
   if no_ode_fit :
      assert job_table == None
   else :
      assert type(job_table) == list
   # END_DEF
   #
   # predict_sample
   predict_sample = not no_ode_fit
   #
   # all_table
   connection = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   all_table  = dict()
   for name in [
      'option_all',
      'split_reference',
      'mulcov_freeze',
      'cov_reference',
   ] :
      all_table[name] =  dismod_at.get_table_dict(connection, name)
   connection.close()
   #
   # root_database
   root_database      = None
   for row in all_table['option_all'] :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
   assert root_database != None
   #
   # shift_prior_std_factor, shift_prior_dage, shift_prior_dtime
   shift_prior_std_factor = 1.0
   shift_prior_dage       = True
   shift_prior_dtime      = True
   for row in all_table['option_all'] :
      if row['option_name'] == 'shift_prior_std_factor' :
         shift_prior_std_factor = float( row['option_value'] )
      if row['option_name'] == 'shift_prior_dage' :
         shift_prior_dage = row['option_value'].strip()
         if shift_prior_dage not in [ 'true', 'false' ] :
            msg  = f'option_all table: shift_prior_dage = {shift_prior_dage} '
            msg += 'is not true or false'
            assert False, msg
         shift_prior_dage = shift_prior_dage == 'true'
      if row['option_name'] == 'shift_prior_dtime' :
         shift_prior_dtime = row['option_value'].strip()
         if shift_prior_dtime not in [ 'true', 'false' ] :
            msg  = f'option_all table: shift_prior_dtime = {shift_prior_dtime} '
            msg += 'is not true or false'
            assert False, msg
         shift_prior_dtime = shift_prior_dtime == 'true'
   #
   # shift_prior_std_factor_mulcov
   shift_prior_std_factor_mulcov = shift_prior_std_factor
   for row in all_table['option_all'] :
      if row['option_name'] == 'shift_prior_std_factor_mulcov' :
         shift_prior_std_factor_mulcov = float( row['option_value'] )
   #
   # no_ode_ignore
   no_ode_ignore = ''
   for row in all_table['option_all'] :
      if row['option_name'] == 'no_ode_ignore' :
         no_ode_ignore = row['option_value'].strip()
   #
   # freeze_type
   freeze_type = 'mean'
   for row in all_table['option_all'] :
      if row['option_name'] == 'freeze_type' :
         freeze_type = row['option_value'].strip()
   if freeze_type not in [ 'mean', 'posterior' ] :
      msg  = f'option_all table: freeze_type = {freeze_type} '
      msg += 'is not "mean" or "posterior"'
      assert False, msg
   #
   # fit_table
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   fit_table  = dict()
   for name in [
      'age',
      'c_shift_avgint',
      'c_shift_predict_fit_var',
      'covariate',
      'density',
      'fit_var',
      'integrand',
      'mulcov',
      'node',
      'option',
      'prior',
      'rate',
      'smooth',
      'smooth_grid',
      'time',
      'var',
   ] :
      fit_table[name] = fit_or_root.get_table(name)
   if predict_sample :
      for name in [ 'c_shift_predict_sample', 'sample' ] :
         fit_table[name] = fit_or_root.get_table(name)
   fit_or_root.close()
   #
   # age_id_next_list
   age_id_next_list = get_age_id_next_list(
      fit_table['smooth'], fit_table['smooth_grid'], fit_table['age']
   )
   #
   # time_id_next_list
   time_id_next_list = get_time_id_next_list(
      fit_table['smooth'], fit_table['smooth_grid'], fit_table['time']
   )
   #
   # name_rate2integrand
   name_rate2integrand = {
      'pini'  : 'prevalence',
      'iota'  : 'Sincidence',
      'rho'   : 'remission',
      'chi'   : 'mtexcess',
   }
   #
   # fit_split_reference_id, split_covariate_id
   cov_info = at_cascade.get_cov_info(
      all_table['option_all'],
      fit_table['covariate'],
      all_table['split_reference']
   )
   if len(all_table['split_reference']) == 0 :
      fit_split_reference_id = None
      split_covaraite_id     = None
   else :
      fit_split_reference_id = cov_info['split_reference_id']
      split_covariate_id     = cov_info['split_covariate_id']
   #
   # fit_fit_var
   fit_fit_var = dict()
   for predict_row in fit_table['c_shift_predict_fit_var'] :
      avgint_id          = predict_row['avgint_id']
      avgint_row         = fit_table['c_shift_avgint'][avgint_id]
      integrand_id       = avgint_row['integrand_id']
      node_id            = avgint_row['node_id']
      age_id             = avgint_row['c_age_id']
      time_id            = avgint_row['c_time_id']
      split_id           = avgint_row['c_split_reference_id']
      key           = (integrand_id, node_id, split_id, age_id, time_id)
      assert not key in fit_fit_var
      fit_fit_var[key] = predict_row['avg_integrand']
   #
   # fit_sample
   fit_sample = dict()
   if predict_sample :
      for predict_row in fit_table['c_shift_predict_sample'] :
         avgint_id          = predict_row['avgint_id']
         avgint_row         = fit_table['c_shift_avgint'][avgint_id]
         integrand_id       = avgint_row['integrand_id']
         node_id            = avgint_row['node_id']
         age_id             = avgint_row['c_age_id']
         time_id            = avgint_row['c_time_id']
         split_id           = avgint_row['c_split_reference_id']
         key           = (integrand_id, node_id, split_id, age_id, time_id)
         if not key in fit_sample :
            fit_sample[key] = list()
         fit_sample[key].append( predict_row['avg_integrand'] )
   #
   # fit_node_name
   fit_node_name = None
   for row in fit_table['option'] :
      assert row['option_name'] != 'fit_node_id'
      if row['option_name'] == 'parent_node_name' :
         fit_node_name = row['option_value']
   assert fit_node_name is not None
   #
   # fit_node_id
   fit_node_id = at_cascade.table_name2id(
      fit_table['node'], 'node', fit_node_name
   )
   for shift_name in shift_databases :
      # ---------------------------------------------------------------------
      # create shift_databases[shift_name]
      # ---------------------------------------------------------------------
      #
      # shift_table
      shift_table = dict()
      for name in [
         'covariate',
         'mulcov',
         'option',
         'rate',
      ] :
         if name not in at_cascade.constant_table_list :
            shift_table[name] = copy.deepcopy(fit_table[name])
      shift_table['prior']       = list()
      shift_table['smooth']      = list()
      shift_table['smooth_grid'] = list()
      shift_table['nslist']      = list()
      shift_table['nslist_pair'] = list()
      #
      # shift_node_name, shift_split_reference_name
      shift_node_name            = None
      shift_split_reference_name = None
      if shift_name.find('/') < 0 :
         for row in all_table['split_reference'] :
            if row['split_reference_name'] == shift_name :
               if shift_node_name != None :
                  msg  = f'{shift_name} is both a split_reference_name '
                  msg += 'and a node_name'
                  assert False, msg
               shift_split_reference_name = shift_name
               shift_node_name            = fit_node_name
         if shift_split_reference_name == None :
            shift_node_name = shift_name
      else :
         shift_split_reference_name = shift_name.split('/')[0]
         shift_node_name            = shift_name.split('/')[1]
      #
      # shift_split_reference_id
      if shift_split_reference_name == None :
         shift_split_reference_id = fit_split_reference_id
      else :
         shift_split_reference_id = at_cascade.table_name2id(
            all_table['split_reference'],
            'split_reference',
            shift_split_reference_name
         )
      #
      # shift_node_id
      shift_node_id  = at_cascade.table_name2id(
         fit_table['node'], 'node', shift_node_name
      )
      #
      # mulcov_freeze_dict
      if no_ode_fit :
         mulcov_freeze_dict = dict()
      else :
         mulcov_freeze_dict = at_cascade.get_freeze_dict(
            fit_table['node'],
            fit_node_id,
            fit_split_reference_id,
            all_table['mulcov_freeze'],
         )
      #
      # shift_database     = fit_database
      shift_database = shift_databases[shift_name]
      shutil.copyfile(fit_database, shift_database)
      #
      # shift_table['option']
      # Set value for parent_node_name and other_database
      for row in shift_table['option'] :
         if row['option_name'] == 'parent_node_name' :
            row['option_value'] = shift_node_name
         if row['option_name'] == 'other_database' :
            if os.path.isabs( root_database ) :
               row['option_value'] = root_database
            else :
               dirname       = os.path.dirname( shift_database )
               relative_path = os.path.relpath( root_database, dirname)
               row['option_value'] = str( relative_path )
         #
      # cov_reference_list
      node_id = fit_table['node'][shift_node_id]['parent']
      assert shift_node_id == fit_node_id or node_id == fit_node_id
      n_covariate = len( fit_table['covariate'] )
      cov_reference_list = get_cov_reference_list(
         n_covariate,
         all_table['cov_reference'],
         shift_node_id,
         shift_split_reference_id
      )
      #
      # shift_table['covariate']
      # set relative covariate values so correspond to shift node
      for (covariate_id, shift_row) in enumerate(shift_table['covariate']) :
            shift_row['reference'] =  cov_reference_list[covariate_id]

      #
      # shift_table['covariate']
      # set shift covaraite value
      if shift_split_reference_id is not None :
         split_row  = all_table['split_reference'][shift_split_reference_id]
         reference  = split_row['split_reference_value']
         shift_row  = shift_table['covariate'][split_covariate_id]
         shift_row['reference'] = reference
      #
      # --------------------------------------------------------------------
      # shift_table['mulcov']
      # and corresponding entries in
      # smooth, smooth_grid, and prior
      for (mulcov_id, shift_mulcov_row) in enumerate(shift_table['mulcov']) :
         assert shift_mulcov_row['subgroup_smooth_id'] is None
         #
         # fit_smooth_id
         fit_smooth_id = shift_mulcov_row['group_smooth_id']
         if not fit_smooth_id is None :
            #
            # integrand_id
            # This is the integrand_id corresponding to this mulcov value.
            # The integrand_id that is affected by the mulcov is called
            # affected_id below.
            name         = 'mulcov_' + str(mulcov_id)
            integrand_id = at_cascade.table_name2id(
               fit_table['integrand'], 'integrand', name
            )
            #
            # smooth_row
            smooth_row = fit_table['smooth'][fit_smooth_id]
            smooth_row = copy.copy(smooth_row)
            assert smooth_row['mulstd_value_prior_id'] is None
            assert smooth_row['mulstd_dage_prior_id']  is None
            assert smooth_row['mulstd_dtime_prior_id'] is None
            #
            # shift_table['smooth'], shift_smooth_id
            shift_smooth_id = len(shift_table['smooth'])
            smooth_row['smooth_name'] += f'_{shift_smooth_id}'
            shift_table['smooth'].append(smooth_row)
            #
            # change shift_table['mulcov'] to use the new smoothing
            shift_mulcov_row['group_smooth_id'] = shift_smooth_id
            #
            # freeze
            assert len(mulcov_freeze_dict) == 0 or not no_ode_fit
            if mulcov_id in mulcov_freeze_dict :
               if freeze_type == 'mean' :
                  freeze = 'mean'
               elif mulcov_freeze_dict[mulcov_id] == 'posterior' :
                  freeze = 'no'
               else :
                  assert mulcov_freeze_dict[mulcov_id] == 'prior'
                  freeze = 'prior'
            else :
               freeze = 'no'
            #
            # copy_row
            copy_row = freeze == 'prior'
            if no_ode_fit :
               mulcov_type = shift_mulcov_row['mulcov_type']
               if mulcov_type == 'rate_value' :
                  rate_id   = shift_mulcov_row['rate_id']
                  rate_name = shift_table['rate'][rate_id]['rate_name']
                  if rate_name in no_ode_ignore.split() :
                     copy_row = True
               if mulcov_type == 'meas_value' :
                  affected_id    = shift_mulcov_row['integrand_id']
                  integrand_row  = fit_table['integrand'][affected_id]
                  integrand_name = integrand_row['integrand_name']
                  if integrand_name in no_ode_ignore.split() :
                     copy_row = True
            #
            # shift_table['smooth_grid']
            # add rows for this smoothing
            node_id  = None
            split_id = None
            for fit_grid_row in fit_table['smooth_grid'] :
               if fit_grid_row['smooth_id'] == fit_smooth_id :
                  add_shift_grid_row(
                     fit_fit_var,
                     fit_sample,
                     fit_table,
                     shift_table,
                     fit_grid_row,
                     integrand_id,
                     node_id,
                     split_id,
                     shift_prior_std_factor_mulcov,
                     shift_prior_dage,
                     shift_prior_dtime,
                     freeze,
                     copy_row,
                     age_id_next_list[fit_smooth_id],
                     time_id_next_list[fit_smooth_id],
                  )

      # --------------------------------------------------------------------
      # shift_table['rate']
      # and corresponding entries in the following child tables:
      # smooth, smooth_grid, and prior
      for shift_rate_row in shift_table['rate'] :
         # rate_name
         rate_name        = shift_rate_row['rate_name']
         # ----------------------------------------------------------------
         # fit_smooth_id
         fit_smooth_id = None
         if rate_name in name_rate2integrand :
            assert shift_rate_row['child_nslist_id'] is None
            fit_smooth_id = shift_rate_row['parent_smooth_id']
         else :
            # proper priors for omega are set by omega_constraint routine
            assert rate_name == 'omega'
            shift_rate_row['parent_smooth_id'] = None
            shift_rate_row['child_smooth_id']  = None
            shift_rate_row['child_nslist_id']  = None
         if not fit_smooth_id is None :
            #
            # integrand_id
            # only check for integrands that are used
            integrand_name  = name_rate2integrand[rate_name]
            integrand_id = at_cascade.table_name2id(
               fit_table['integrand'], 'integrand', integrand_name
            )
            #
            # smooth_row
            smooth_row = fit_table['smooth'][fit_smooth_id]
            smooth_row = copy.copy(smooth_row)
            assert smooth_row['mulstd_value_prior_id'] is None
            assert smooth_row['mulstd_dage_prior_id']  is None
            assert smooth_row['mulstd_dtime_prior_id'] is None
            #
            # : shift_table['smooth'], shift_smooth_id
            shift_smooth_id = len(shift_table['smooth'])
            smooth_row['smooth_name'] += f'_{shift_smooth_id}'
            shift_table['smooth'].append(smooth_row)
            #
            # shift_table['rate']
            # use the new smoothing for this rate
            shift_rate_row['parent_smooth_id'] = shift_smooth_id
            #
            # freeze
            freeze = 'no'
            #
            # copy_row
            copy_row = False
            if no_ode_fit :
               if rate_name in no_ode_ignore.split() :
                  copy_row = True
            #
            # shift_table['smooth_grid']
            # add rows for this smoothing
            for fit_grid_row in fit_table['smooth_grid'] :
               if fit_grid_row['smooth_id'] == fit_smooth_id :
                  add_shift_grid_row(
                     fit_fit_var,
                     fit_sample,
                     fit_table,
                     shift_table,
                     fit_grid_row,
                     integrand_id,
                     shift_node_id,
                     shift_split_reference_id,
                     shift_prior_std_factor,
                     shift_prior_dage,
                     shift_prior_dtime,
                     freeze,
                     copy_row,
                     age_id_next_list[fit_smooth_id],
                     time_id_next_list[fit_smooth_id],
                  )
         # ----------------------------------------------------------------
         # fit_smooth_id
         fit_smooth_id = None
         if rate_name in name_rate2integrand :
            fit_smooth_id = shift_rate_row['child_smooth_id']
         if not fit_smooth_id is None :
            #
            smooth_row = fit_table['smooth'][fit_smooth_id]
            smooth_row = copy.copy(smooth_row)
            #
            assert smooth_row['mulstd_value_prior_id'] is None
            assert smooth_row['mulstd_dage_prior_id']  is None
            assert smooth_row['mulstd_dtime_prior_id'] is None
            if rate_name == 'pini' :
               assert smooth_row['n_age'] == 1
            #
            # update: shift_table['smooth']
            # for case where its is the parent
            shift_smooth_id = len(shift_table['smooth'])
            smooth_row['smooth_name'] += f'_{shift_smooth_id}'
            shift_table['smooth'].append(smooth_row)
            #
            # change shift_table['rate'] to use the new smoothing
            shift_rate_row['child_smooth_id'] = shift_smooth_id
            #
            # add rows for this smoothing to shift_table['smooth_grid']
            for fit_grid_row in fit_table['smooth_grid'] :
               if fit_grid_row['smooth_id'] == fit_smooth_id :
                  #
                  # update: shift_table['smooth_grid']
                  shift_grid_row = copy.copy( fit_grid_row )
                  #
                  for ty in [
                     'value_prior_id', 'dage_prior_id', 'dtime_prior_id'
                         ] :
                     prior_id  = fit_grid_row[ty]
                     if prior_id is None :
                        shift_grid_row[ty] = None
                     else :
                        prior_row = fit_table['prior'][prior_id]
                        prior_row = copy.copy(prior_row)
                        prior_id  = len( shift_table['prior'] )
                        shift_table['prior'].append( prior_row )
                        add_index_to_name(
                           shift_table['prior'], 'prior_name'
                        )
                        shift_grid_row[ty] = prior_id
                  shift_grid_row['smooth_id']      = shift_smooth_id
                  shift_table['smooth_grid'].append( shift_grid_row )
      #
      # shift_connection
      new        = False
      shift_connection = dismod_at.create_connection(shift_database, new)
      #
      # replace shift_table
      for name in shift_table :
         dismod_at.replace_table(
            shift_connection, name, shift_table[name]
         )
      #
      # empty_avgint_table
      at_cascade.empty_avgint_table(shift_connection)
      #
      # drop the following tables:
      # log, c_shift_avgint, c_shift_predict_sample, c_shift_predict_fit_var
      drop_list = [ 'log', 'c_shift_avgint' , 'c_shift_predict_fit_var' ]
      if predict_sample :
         drop_list.append(  'c_shift_predict_sample' )
      for table_name in drop_list :
         command  = f'DROP TABLE {table_name}'
         dismod_at.sql_command(shift_connection, command)
      #
      # shift_connection
      shift_connection.close()
      #
      # shift_database
      at_cascade.omega_constraint(all_node_database, shift_database)



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_var_id}
{xrst_spell
  dage
  dtime
  meas
  mulstd
}

Map Var Table Values to its Primary Key
#######################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

var_table
*********
This is a ``list`` of ``dict`` representation of the
a dismod_at var table.

var_type
********
This is a ``str`` that specifies the type of variable
for the corresponding row of the var table.
It must have one of the values in the table below
(and cannot be ``None``).


Other Arguments
***************
All of the arguments, other than *var_type*, are ``int`` values.
Depending on the value of *var_type* only certain of the
other arguments matter:

.. csv-table:: **Other Arguments That Matter by Variable Type**
   :header: var_type,     Other Arguments
   :widths: 30, 70

   mulstd_value,          *smooth_id*
   mulstd_dage,           *smooth_id*
   mulstd_dtime,          *smooth_id*
   rate,                  "age_id, time_id, node_id, rate_id"
   mulcov_rate_value,     "age_id, time_id, mulcov_id, group_id, subgroup_id"
   mulcov_meas_value,     "age_id, time_id, mulcov_id, group_id, subgroup_id"
   mulcov_meas_noise,     "age_id, time_id, mulcov_id, group_id"


{xrst_end get_var_id}
'''
# BEGIN_DEF
# at_cascade.get_var_id
def get_var_id(
   var_table          ,
   var_type           ,
   smooth_id    = None,
   age_id       = None,
   time_id      = None,
   node_id      = None,
   rate_id      = None,
   mulcov_id    = None,
   group_id     = None,
   subgroup_id  = None,
) :
   assert type(var_table) == list
   assert type(var_type)  == str
   # END_DEF
   # var_id
   var_id       = None
   #
   # double_match
   double_match = False
   #
   # mulstd_value, mulstd_dage, mulstd_dtime
   if var_type in [ 'mulstd_value', 'mulstd_dage', 'mulstd_dtime' ] :
      for (row_id, row) in enumerate(var_table) :
         match = True
         match = match and row['var_type']  == var_type
         match = match and row['smooth_id'] == smooth_id
         if match :
            # double_match
            if not var_id is None :
               double_match = True
            # var_id
            var_id = row_id
   # rate
   elif var_type == 'rate' :
      for (row_id, row) in enumerate(var_table) :
         match = True
         match = match and row['var_type']  == var_type
         match = match and row['age_id']    == age_id
         match = match and row['time_id']   == time_id
         match = match and row['node_id']   == node_id
         match = match and row['rate_id']   == rate_id
         if match :
            # double_match
            if not var_id is None :
               double_match = True
            # var_id
            var_id = row_id
   # mulcov_rate_value, mulcov_meas_value
   elif var_type in [ 'mulcov_rate_value', 'mulcov_meas_value' ] :
      for (row_id, row) in enumerate(var_table) :
         match = True
         match = match and row['var_type']    == var_type
         match = match and row['age_id']      == age_id
         match = match and row['time_id']     == time_id
         match = match and row['mulcov_id']   == mulcov_id
         match = match and row['group_id']    == group_id
         match = match and row['subgroup_id'] == subgroup_id
         if match :
            # double_match
            if not var_id is None :
               double_match = True
            # var_id
            var_id = row_id
   # mulcov_meas_noise
   else :
      assert var_type == 'mulcov_meas_noise'
      for (row_id, row) in enumerate(var_table) :
         match = True
         match = match and row['var_type']    == var_type
         match = match and row['age_id']      == age_id
         match = match and row['time_id']     == time_id
         match = match and row['mulcov_id']   == mulcov_id
         match = match and row['group_id']    == group_id
         if match :
            # double_match
            if not var_id is None :
               double_match = True
            # var_id
            var_id = row_id
   # double_match
   if double_match :
      for row in var_table :
         print(row)
      msg  = 'get_var_id: Something is wrong with this var_table'
      assert False, msg
   #
   # BEGIN_RETURN
   # ...
   assert type(var_id) == int
   return var_id
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin csv.check_table}

Check Columns in a CSV File
###########################

Prototype
*********
{xrst_literal
   BEGIN_DEF
   END_DEF
}

file_name
*********
is the name of the name of the csv file corresponding to the table.

table
*****
is the table corresponding to the csv file; i.e.,
{xrst_code py}
table  = at_cascade.csv.read(file_name)
{xrst_code}

{xrst_end csv.check_table}
'''
#
# BEGIN_DEF
# at_cascade.csv.check_table
def check_table(file_name, table) :
   assert type(file_name) == str
   assert type(table) == list
   if len(table) > 0 :
      assert type( table[0] ) == dict
   # END_DEF
   #
   # empty table case
   if len(table) == 0 :
      return
   #
   # table_name
   if '/' in file_name :
      index      = file_name.rfind('/')
      table_name = file_name[index+1 :]
   else :
      table_name = file_name
   index      = table_name.rfind('.csv')
   table_name = table_name[: index]
   #
   # table2keys
   table2keys = {
      #
      # simulate tables
      'option_sim'     : [ 'name', 'value' ],
      'node'           : [ 'node_name', 'parent_name' ],
      'no_effect_rate' : [ 'rate_name', 'age', 'time', 'rate_truth' ],
      'covariate' : [
            'node_name',
            'sex',
            'age',
            'time',
            'omega',
      ],
      'multiplier_sim' : [
         'multiplier_id',
         'rate_name',
         'covariate_or_sex',
         'multiplier_truth',
      ],
      'simulate' : [
         'simulate_id',
         'integrand_name',
         'node_name',
         'sex',
         'age_lower',
         'age_upper',
         'time_lower',
         'time_upper',
         'meas_std_min',
         'meas_std_cv',
      ],
      #
      # fit table that are not simulate tables
      'option_fit'        : [ 'name', 'value' ],
      'fit_goal'          : [ 'node_name' ],
      'predict_integrand' : [ 'integrand_name' ],
      'child_rate'        : [ 'rate_name', 'value_prior' ],
      'prior' : [
         'name',
         'lower',
         'upper',
         'mean',
         'std',
         'density',
      ],
      'parent_rate' : [
         'rate_name',
         'age',
         'time',
         'value_prior',
         'dage_prior',
         'dtime_prior',
         'const_value',
      ],
      'mulcov' : [
         'covariate',
         'type',
         'effected',
         'value_prior',
         'const_value',
      ],
      'data_in' : [
         'data_id',
         'integrand_name',
         'node_name',
         'sex',
         'age_lower',
         'age_upper',
         'time_lower',
         'time_upper',
         'meas_value',
         'meas_std',
         'hold_out',
      ],
   }
   #
   # key
   keys = table2keys[table_name]
   for key in keys :
         if key not in table[0] :
            msg       = f'The column name {key} does not appear in this file\n'
            msg      += f'{file_name}'
            assert False, msg
   #
   return
